#!/usr/bin/env python
"""Basic rdfvalue tests."""

import datetime
import time

from absl import app

from grr_response_core.lib import rdfvalue
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr.test_lib import test_lib

BILLION = 1000000000


class RDFBytesTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.RDFBytes

  def GenerateSample(self, number=0):
    return rdfvalue.RDFBytes(b"\x00hello%s\x01" % str(number).encode("ascii"))


class RDFStringTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.RDFString

  def GenerateSample(self, number=0):
    return rdfvalue.RDFString("Grüezi %s" % number)

  def testRDFStringGetItem(self):
    rdfstring = rdfvalue.RDFString("123456789")
    self.assertEqual(rdfstring[3], "4")
    self.assertEqual(rdfstring[-3:], "789")
    self.assertEqual(rdfstring[3:-3], "456")


class RDFIntegerTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.RDFInteger

  def GenerateSample(self, number=0):
    return rdfvalue.RDFInteger(number)

  def testComparableToPrimiviteInts(self):
    self.assertEqual(rdfvalue.RDFInteger(10), 10)
    self.assertGreater(rdfvalue.RDFInteger(10), 5)
    self.assertGreater(15, rdfvalue.RDFInteger(10))
    self.assertLess(rdfvalue.RDFInteger(10), 15)
    self.assertLess(5, rdfvalue.RDFInteger(10))

  def testDividesAndIsDividableByPrimitiveInts(self):
    self.assertEqual(rdfvalue.RDFInteger(10) // 5, 2)

  def testMultipliesAndIsMultipliedByByPrimitive(self):
    self.assertEqual(rdfvalue.RDFInteger(10) * 10, 100)
    self.assertEqual(10 * rdfvalue.RDFInteger(10), 100)

  def testUsableInBitwiseOr(self):
    def TestOr(val1, val2, expected):
      self.assertEqual(rdfvalue.RDFInteger(val1) | val2, expected)
      self.assertEqual(val1 | rdfvalue.RDFInteger(val2), expected)

      value = rdfvalue.RDFInteger(val1)
      value |= val2
      self.assertEqual(value, expected)

      value = val1
      value |= rdfvalue.RDFInteger(val2)
      self.assertEqual(value, expected)

    TestOr(True, False, True)
    TestOr(False, True, True)
    TestOr(False, False, False)
    TestOr(True, True, True)

  def testUsableInBitwiseAnd(self):
    def TestAnd(val1, val2, expected):
      self.assertEqual(rdfvalue.RDFInteger(val1) & val2, expected)
      self.assertEqual(val1 & rdfvalue.RDFInteger(val2), expected)

      value = rdfvalue.RDFInteger(val1)
      value &= val2
      self.assertEqual(value, expected)

      value = val1
      value &= rdfvalue.RDFInteger(val2)
      self.assertEqual(value, expected)

    TestAnd(True, False, False)
    TestAnd(False, True, False)
    TestAnd(False, False, False)
    TestAnd(True, True, True)


class DurationSecondsTest(
    rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest
):
  rdfvalue_class = rdfvalue.DurationSeconds

  def GenerateSample(self, number=5):
    return rdfvalue.DurationSeconds("%ds" % number)

  def testStringRepresentationIsTransitive(self):
    t = rdfvalue.DurationSeconds("5 m")
    self.assertEqual(t.ToInt(rdfvalue.SECONDS), 300)
    self.assertEqual(t, rdfvalue.DurationSeconds(300))
    self.assertEqual(str(t), "5 m")

  def testMulNumber(self):
    t = rdfvalue.DurationSeconds("5m")
    t2 = t * 3
    self.assertEqual(t2.ToInt(rdfvalue.SECONDS), 300 * 3)

    # Test rmul
    t2 = 3 * t
    self.assertEqual(t2.ToInt(rdfvalue.SECONDS), 300 * 3)

  def testHashability(self):
    pass  # DurationSeconds does not need to be hashable.


class ByteSizeTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.ByteSize

  def GenerateSample(self, number=5):
    return rdfvalue.ByteSize("%sKib" % number)

  def testParsing(self):
    cases = [
        ("100gb", 100 * 1000**3),
        ("10kib", 10 * 1024),
        ("2.5kb", 2500),
        ("3.25MiB", 3.25 * 1024**2),
        ("3.25 MiB", 3.25 * 1024**2),
        ("12B", 12),
        ("12 B", 12),
    ]

    for string, expected in cases:
      self.assertEqual(expected, rdfvalue.ByteSize(string))


class RDFURNTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.RDFURN

  def GenerateSample(self, number=0):
    return rdfvalue.RDFURN("aff4:/C.12342%s/fs/os/" % number)

  def testRDFURN(self):
    """Test RDFURN handling."""
    # Make a url object
    str_url = "aff4:/hunts/W:AAAAAAAA/Results"
    url = rdfvalue.RDFURN(str_url)
    self.assertEqual(url.Path(), "/hunts/W:AAAAAAAA/Results")
    self.assertEqual(str(url), str_url)
    self.assertEqual(url.scheme, "aff4")

    # Test the Add() function
    url = url.Add("some").Add("path")
    self.assertEqual(url.Path(), "/hunts/W:AAAAAAAA/Results/some/path")
    self.assertEqual(str(url), utils.Join(str_url, "some", "path"))

    # Test that we can handle urns with a '?' and do not interpret them as
    # a delimiter between url and parameter list.
    str_url = "aff4:/C.0000000000000000/fs/os/c/regex.*?]&[+{}--"
    url = rdfvalue.RDFURN(str_url)
    self.assertEqual(url.Path(), str_url[5:])

    # Some more special characters...
    for path in ["aff4:/test/?#asd", "aff4:/test/#asd", "aff4:/test/?#"]:
      self.assertEqual(path, str(rdfvalue.RDFURN(path)))

  def testComparison(self):
    urn = rdfvalue.RDFURN("aff4:/abc/def")
    self.assertEqual(urn, str(urn))
    self.assertEqual(urn, "aff4:/abc/def")
    self.assertEqual(urn, "/abc/def")
    self.assertEqual(urn, "abc/def")

    self.assertIsNotNone(urn)

    string_list = ["abc", "ghi", "def", "mno", "jkl"]
    urn_list = [rdfvalue.RDFURN(s) for s in string_list]

    self.assertEqual(string_list, urn_list)

    # Inequality.
    s = "some_urn"
    s2 = "some_other_urn"
    self.assertEqual(s, rdfvalue.RDFURN(s))
    self.assertFalse(s != rdfvalue.RDFURN(s))  # pylint: disable=g-generic-assert
    self.assertNotEqual(s, rdfvalue.RDFURN(s2))
    self.assertFalse(s == rdfvalue.RDFURN(s2))  # pylint: disable=g-generic-assert

  def testHashing(self):

    m = {}
    urn1 = rdfvalue.RDFURN("aff4:/test1")
    urn2 = rdfvalue.RDFURN("aff4:/test2")

    m[urn1] = 1
    self.assertIn(urn1, m)
    self.assertNotIn(urn2, m)

  def testInitialization(self):
    """Check that we can initialize from common initializers."""

    # Initialize from another instance.
    sample = self.GenerateSample("aff4:/")

    self.CheckRDFValue(self.rdfvalue_class(sample), sample)

  def testSerialization(self, sample=None):
    sample = self.GenerateSample("aff4:/")
    super().testSerialization(sample=sample)


class RDFDatetimeTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.RDFDatetime

  def GenerateSample(self, number=0):
    result = self.rdfvalue_class.FromHumanReadable(
        "2011/11/%02d" % (number + 1)
    )
    return result

  def testTimeZoneConversions(self):
    time_string = "2011-11-01 10:23:00"

    # Human readable strings are assumed to always be in UTC
    # timezone. Initialize from the human readable string.
    date1 = self.rdfvalue_class.FromHumanReadable(time_string)

    self.assertEqual(date1.AsMicrosecondsSinceEpoch(), 1320142980000000)

    self.assertEqual(
        time.strftime(
            "%Y-%m-%d %H:%M:%S", time.gmtime(date1.AsSecondsSinceEpoch())
        ),
        time_string,
    )

    # We always stringify the date in UTC timezone.
    self.assertEqual(str(date1), time_string)

  def testInitFromEmptyString(self):
    with test_lib.FakeTime(1000):
      # Init from an empty string should generate a DateTime object with a zero
      # time.
      date = self.rdfvalue_class.FromSerializedBytes(b"")
      self.assertEqual(int(date), 0)

      self.assertEqual(
          self.rdfvalue_class.Now().AsMicrosecondsSinceEpoch(), 1000 * 1000000
      )

  def testInitFromDatetimeObject(self):
    # Test initializing from a datetime object
    date = datetime.datetime(2015, 6, 17, 5, 22, 3)
    self.assertEqual(self.rdfvalue_class.FromDatetime(date).AsDatetime(), date)
    date = datetime.datetime.utcfromtimestamp(99999)
    self.assertEqual(
        self.rdfvalue_class.FromDatetime(date).AsSecondsSinceEpoch(), 99999
    )
    self.assertEqual(
        self.rdfvalue_class.FromDatetime(date).AsMicrosecondsSinceEpoch(),
        99999000000,
    )
    # Test microsecond support
    date = datetime.datetime(1970, 1, 1, 0, 0, 0, 567)
    self.assertEqual(
        rdfvalue.RDFDatetime.FromDatetime(date).AsMicrosecondsSinceEpoch(), 567
    )

  def testInitFromDateObject(self):
    date = datetime.date(2018, 2, 1)
    self.assertEqual(
        self.rdfvalue_class.FromDate(date),
        self.rdfvalue_class.FromHumanReadable("2018-02-01 00:00:00"),
    )

  def testAddNumber(self):
    date = rdfvalue.RDFDatetime(BILLION)
    self.assertEqual(int(date + 60), BILLION + 60 * 1000000)
    self.assertEqual(int(date + (-10)), BILLION - 10 * 1000000)

  def testSubNumber(self):
    date = rdfvalue.RDFDatetime(BILLION)
    self.assertEqual(int(date - 60), BILLION - 60 * 1000000)
    self.assertEqual(int(date - int(1e12)), BILLION - int(1e12) * 1000000)

  def testIAddNumber(self):
    date = rdfvalue.RDFDatetime(BILLION)
    date += 60
    self.assertEqual(date, BILLION + 60 * 1000000)

    date = rdfvalue.RDFDatetime(BILLION)
    date += -10
    self.assertEqual(date, BILLION - 10 * 1000000)

  def testISubNumber(self):
    date = rdfvalue.RDFDatetime(BILLION)
    date -= 60
    self.assertEqual(date, BILLION - 60 * 1000000)

    date = rdfvalue.RDFDatetime(BILLION)
    date -= int(1e12)
    self.assertEqual(date, BILLION - int(1e12) * 1000000)

  def testAddDuration(self):
    duration = rdfvalue.Duration.FromHumanReadable("12h")
    date = self.rdfvalue_class.FromMicrosecondsSinceEpoch(BILLION)
    self.assertEqual(
        (date + duration).AsMicrosecondsSinceEpoch(),
        BILLION + 12 * 3600 * 1000000,
    )

  def testSubDuration(self):
    duration = rdfvalue.Duration.FromHumanReadable("5m")
    date = self.rdfvalue_class.FromMicrosecondsSinceEpoch(BILLION)
    self.assertEqual(
        (date - duration).AsMicrosecondsSinceEpoch(), BILLION - 5 * 60 * 1000000
    )
    duration = rdfvalue.Duration.FromHumanReadable("1w")
    self.assertEqual(
        (date - duration).AsMicrosecondsSinceEpoch(),
        BILLION - 7 * 24 * 3600 * 1000000,
    )

  def testIAddDuration(self):
    date = self.rdfvalue_class.FromMicrosecondsSinceEpoch(BILLION)
    date += rdfvalue.Duration.FromHumanReadable("12h")
    self.assertEqual(
        date.AsMicrosecondsSinceEpoch(), BILLION + 12 * 3600 * 1000000
    )

  def testISubDuration(self):
    date = self.rdfvalue_class.FromMicrosecondsSinceEpoch(BILLION)
    date -= rdfvalue.Duration.FromHumanReadable("5m")
    self.assertEqual(
        date.AsMicrosecondsSinceEpoch(), BILLION - 5 * 60 * 1000000
    )

    date = self.rdfvalue_class.FromMicrosecondsSinceEpoch(BILLION)
    date -= rdfvalue.Duration.FromHumanReadable("1w")
    self.assertEqual(
        date.AsMicrosecondsSinceEpoch(), BILLION - 7 * 24 * 3600 * 1000000
    )


class RDFDatetimeSecondsTest(RDFDatetimeTest):
  rdfvalue_class = rdfvalue.RDFDatetimeSeconds


class HashDigestTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdfvalue.HashDigest

  def GenerateSample(self, number=0):
    return rdfvalue.HashDigest(
        b"\xca\x97\x81\x12\xca\x1b\xbd\xca\xfa\xc21\xb3"
        b"\x9a#\xdcM\xa7\x86\xef\xf8\x14|Nr\xb9\x80w\x85"
        b"\xaf\xeeH\xbb%s"
        % str(number).encode("ascii")
    )

  def testEqNeq(self):
    binary_digest = (
        b"\xca\x97\x81\x12\xca\x1b\xbd\xca\xfa\xc21\xb3"
        b"\x9a#\xdcM\xa7\x86\xef\xf8\x14|Nr\xb9\x80w\x85"
        b"\xaf\xeeH\xbb"
    )
    sample = rdfvalue.HashDigest(binary_digest)
    hex_digest = (
        "ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb"
    )
    self.assertEqual(str(sample), hex_digest)
    self.assertEqual(sample.SerializeToBytes(), binary_digest)
    self.assertNotEqual(sample.SerializeToBytes(), b"\xaa\xbb")
    self.assertNotEqual(str(sample), "deadbeef")


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Client actions requests and responses."""

from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import protodict as rdf_protodict
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


class EchoRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.EchoRequest


class ExecuteBinaryRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecuteBinaryRequest
  rdf_deps = [
      rdf_crypto.SignedBlob,
  ]


class ExecuteBinaryResponse(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecuteBinaryResponse


class ExecutePythonRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecutePythonRequest
  rdf_deps = [
      rdf_protodict.Dict,
      rdf_crypto.SignedBlob,
  ]


class ExecutePythonResponse(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecutePythonResponse


class ExecuteRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecuteRequest


class ExecuteResponse(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ExecuteResponse
  rdf_deps = [
      ExecuteRequest,
  ]


class Iterator(rdf_structs.RDFProtoStruct):
  """An Iterated client action is one which can be resumed on the client."""

  protobuf = jobs_pb2.Iterator
  rdf_deps = [
      rdf_protodict.Dict,
  ]


class ListDirRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ListDirRequest
  rdf_deps = [
      Iterator,
      rdf_paths.PathSpec,
  ]


class GetFileStatRequest(rdf_structs.RDFProtoStruct):

  protobuf = jobs_pb2.GetFileStatRequest
  rdf_deps = [
      rdf_paths.PathSpec,
  ]


class FingerprintTuple(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.FingerprintTuple


class FingerprintRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.FingerprintRequest
  rdf_deps = [
      FingerprintTuple,
      rdf_paths.PathSpec,
  ]

  def AddRequest(self, *args, **kw):
    self.tuples.Append(*args, **kw)


class FingerprintResponse(rdf_structs.RDFProtoStruct):
  """Proto containing dicts with hashes."""

  protobuf = jobs_pb2.FingerprintResponse
  rdf_deps = [
      rdf_protodict.Dict,
      rdf_crypto.Hash,
      rdf_paths.PathSpec,
  ]

  def GetFingerprint(self, name):
    """Gets the first fingerprint type from the protobuf."""
    for result in self.results:
      if result.GetItem("name") == name:
        return result


class WMIRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.WMIRequest


class StatFSRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.StatFSRequest


class ListNetworkConnectionsArgs(rdf_structs.RDFProtoStruct):
  """Args for the ListNetworkConnections client action."""

  protobuf = flows_pb2.ListNetworkConnectionsArgs



#!/usr/bin/env python
"""A module for working with extended file stat collection.

This module will try to collect as detailed stat information as possible
depending on platform capabilities (e.g. on Linux it will use `statx` [1] call.

[1]: https://www.man7.org/linux/man-pages/man2/statx.2.html
"""

import ctypes
import functools
import operator
import os
import platform
from typing import NamedTuple


# Indicates whether the call also collects information about the birth time.
BTIME_SUPPORT: bool


# TODO(hanuszczak): Migrate to data classes on support for 3.7 is available.
class Result(NamedTuple):
  """A result of extended stat collection."""

  # A bitmask with extra file attributes.
  attributes: int
  # A number of hard links.
  nlink: int
  # A user identifier of the owner.
  uid: int
  # A group identifier of the owner.
  gid: int
  # A bitmask indicating the file and mode of the file.
  mode: int
  # An inode number of the file.
  ino: int
  # The total size (in bytes) of the file.
  size: int

  # Last access time (in nanoseconds since epoch).
  atime_ns: int
  # Last access time (in nanoseconds since epoch).
  btime_ns: int
  # Last access time (in nanoseconds since epoch).
  ctime_ns: int
  # Last access time (in nanoseconds since epoch).
  mtime_ns: int

  # The device identifier (if file represents a device).
  rdev: int
  # The device identifier of the filesystem the file resides on.
  dev: int


def Get(path: bytes) -> Result:
  """Collects detailed stat information about the path.

  Args:
    path: A path to the file for which the information should be retrieved.

  Returns:
    An object with detailed start information.
  """
  return _GetImpl(path)


class _StatxTimestampStruct(ctypes.Structure):
  """A low-level definition of Linux's stat timestamp type."""

  # https://elixir.bootlin.com/linux/v5.6/source/include/uapi/linux/stat.h
  _fields_ = [
      ("tv_sec", ctypes.c_int64),
      ("tv_nsec", ctypes.c_uint32),
      ("__reserved", ctypes.c_int32),
  ]

  @property
  def nanos(self):
    """A number of nanoseconds since epoch the timestamp represents."""
    return self.tv_sec * 10**9 + self.tv_nsec


class _StatxStruct(ctypes.Structure):
  """A low-level definition of Linux's stat object type."""

  # https://elixir.bootlin.com/linux/v5.6/source/include/uapi/linux/stat.h
  _fields_ = [
      ("stx_mask", ctypes.c_uint32),
      ("stx_blksize", ctypes.c_uint32),
      ("stx_attributes", ctypes.c_uint64),
      ("stx_nlink", ctypes.c_uint32),
      ("stx_uid", ctypes.c_uint32),
      ("stx_gid", ctypes.c_uint32),
      ("stx_mode", ctypes.c_uint16),
      ("__spare0", ctypes.c_uint16 * 1),
      ("stx_ino", ctypes.c_uint64),
      ("stx_size", ctypes.c_uint64),
      ("stx_blocks", ctypes.c_uint64),
      ("stx_attributes_mask", ctypes.c_uint64),
      # File timestamps.
      ("stx_atime", _StatxTimestampStruct),
      ("stx_btime", _StatxTimestampStruct),
      ("stx_ctime", _StatxTimestampStruct),
      ("stx_mtime", _StatxTimestampStruct),
      # Device identifier (if the file represents a device).
      ("stx_rdev_major", ctypes.c_uint32),
      ("stx_rdev_minor", ctypes.c_uint32),
      # Device identifier of the filesystem the file resides on.
      ("stx_dev_major", ctypes.c_uint32),
      ("stx_dev_minor", ctypes.c_uint32),
      # Spare space for future extensions.
      ("__spare2", ctypes.c_uint64 * 14),
  ]

  @property
  def rdev(self) -> int:
    """Device identifier (if the file represents a device)."""
    # https://elixir.bootlin.com/linux/v5.6/source/tools/include/nolibc/nolibc.h
    return ((self.stx_rdev_major & 0xFFF) << 8) | (self.stx_rdev_minor & 0xFF)

  @property
  def dev(self) -> int:
    """Device identifier of the filesystem the file resides on."""
    # https://elixir.bootlin.com/linux/v5.6/source/tools/include/nolibc/nolibc.h
    return ((self.stx_dev_major & 0xFFF) << 8) | (self.stx_dev_minor & 0xFF)


# https://elixir.bootlin.com/linux/v3.4/source/include/linux/fcntl.h
_AT_SYMLINK_NOFOLLOW = 0x100
_AT_STATX_SYNC_AS_STAT = 0x0000

# https://elixir.bootlin.com/linux/v5.8/source/include/uapi/linux/stat.h
_STATX_MODE = 0x00000002
_STATX_NLINK = 0x00000004
_STATX_UID = 0x00000008
_STATX_GID = 0x00000010
_STATX_ATIME = 0x00000020
_STATX_BTIME = 0x00000800
_STATX_MTIME = 0x00000040
_STATX_CTIME = 0x00000080
_STATX_INO = 0x00000100
_STATX_SIZE = 0x00000200
_STATX_ALL = functools.reduce(
    operator.__or__,
    [
        _STATX_MODE,
        _STATX_NLINK,
        _STATX_UID,
        _STATX_GID,
        _STATX_ATIME,
        _STATX_BTIME,
        _STATX_MTIME,
        _STATX_CTIME,
        _STATX_INO,
        _STATX_SIZE,
    ],
    0,
)

if platform.system() == "Linux":

  _libc = ctypes.CDLL("libc.so.6")

  try:
    _statx = _libc.statx
  except AttributeError:
    # `statx` is available only since glibc 2.28.
    _statx = None

  if _statx is not None:

    _statx.argtypes = [
        # Input arguments.
        ctypes.c_int,
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.c_uint,
        # Output arguments.
        ctypes.POINTER(_StatxStruct),
    ]
    _statx.restype = ctypes.c_int

    def _GetImplLinuxStatx(path: bytes) -> Result:
      """A Linux-specific stat implementation through `statx`."""
      c_result = _StatxStruct()
      c_status = _statx(
          0,
          path,
          _AT_SYMLINK_NOFOLLOW | _AT_STATX_SYNC_AS_STAT,
          _STATX_ALL,
          ctypes.pointer(c_result),
      )

      if c_status != 0:
        raise OSError(f"Failed to stat '{path}', error code: {c_status}")

      return Result(
          attributes=c_result.stx_attributes,
          nlink=c_result.stx_nlink,
          uid=c_result.stx_uid,
          gid=c_result.stx_gid,
          mode=c_result.stx_mode,
          ino=c_result.stx_ino,
          size=c_result.stx_size,
          atime_ns=c_result.stx_atime.nanos,
          btime_ns=c_result.stx_btime.nanos,
          ctime_ns=c_result.stx_ctime.nanos,
          mtime_ns=c_result.stx_mtime.nanos,
          rdev=c_result.rdev,
          dev=c_result.dev,
      )

    _GetImpl = _GetImplLinuxStatx
    BTIME_SUPPORT = True

  else:

    def _GetImplLinux(path: bytes) -> Result:
      """A generic Linux-specific stat implementation."""
      stat_obj = os.lstat(path)
      return Result(
          attributes=0,  # Not available.
          nlink=stat_obj.st_nlink,
          uid=stat_obj.st_uid,
          gid=stat_obj.st_gid,
          mode=stat_obj.st_mode,
          ino=stat_obj.st_ino,
          size=stat_obj.st_size,
          atime_ns=stat_obj.st_atime_ns,
          btime_ns=0,  # Not available.
          ctime_ns=stat_obj.st_ctime_ns,
          mtime_ns=stat_obj.st_mtime_ns,
          rdev=stat_obj.st_rdev,
          dev=stat_obj.st_dev,
      )

    _GetImpl = _GetImplLinux
    BTIME_SUPPORT = False

elif platform.system() == "Darwin":

  def _GetImplMacos(path: bytes) -> Result:
    """A macOS-specific stat implementation."""
    stat_obj = os.lstat(path)
    # Nanosecond-precision birthtime is not available, with approximate it with
    # the float-precision one.
    st_birthtime_ns = int(stat_obj.st_birthtime * 10**9)  # pytype: disable=attribute-error

    return Result(
        attributes=stat_obj.st_flags,  # pytype: disable=attribute-error
        nlink=stat_obj.st_nlink,
        uid=stat_obj.st_uid,
        gid=stat_obj.st_gid,
        mode=stat_obj.st_mode,
        ino=stat_obj.st_ino,
        size=stat_obj.st_size,
        atime_ns=stat_obj.st_atime_ns,
        btime_ns=st_birthtime_ns,
        ctime_ns=stat_obj.st_ctime_ns,
        mtime_ns=stat_obj.st_mtime_ns,
        rdev=stat_obj.st_rdev,
        dev=stat_obj.st_dev,
    )

  _GetImpl = _GetImplMacos
  BTIME_SUPPORT = True

elif platform.system() == "Windows":

  def _GetImplWindows(path: bytes) -> Result:
    """A Windows-specific stat implementation."""
    stat_obj = os.lstat(path)

    # pylint: disable=line-too-long
    # On Windows, the `st_ctime` field is the file birth time [1], so we just
    # copy this value both to `btime` and `ctime`.
    #
    # [1]: https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/stat-functions
    # pylint: enable=line-too-long

    return Result(
        attributes=stat_obj.st_file_attributes,  # pytype: disable=attribute-error
        nlink=stat_obj.st_nlink,
        uid=stat_obj.st_uid,
        gid=stat_obj.st_gid,
        mode=stat_obj.st_mode,
        ino=stat_obj.st_ino,
        size=stat_obj.st_size,
        atime_ns=stat_obj.st_atime_ns,
        btime_ns=stat_obj.st_ctime_ns,
        ctime_ns=stat_obj.st_ctime_ns,
        mtime_ns=stat_obj.st_mtime_ns,
        rdev=0,  # Not available.
        dev=stat_obj.st_dev,
    )

  _GetImpl = _GetImplWindows
  BTIME_SUPPORT = True



#!/usr/bin/env python
import io

from absl.testing import absltest

from grr_response_core.lib.util import chunked


class ReadWriteTest(absltest.TestCase):

  def testSingleChunk(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"foo")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(chunked.Read(buf), b"foo")
    self.assertIsNone(chunked.Read(buf))

  def testMultipleChunks(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"foo")
    chunked.Write(buf, b"bar")
    chunked.Write(buf, b"baz")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(chunked.Read(buf), b"foo")
    self.assertEqual(chunked.Read(buf), b"bar")
    self.assertEqual(chunked.Read(buf), b"baz")
    self.assertIsNone(chunked.Read(buf))

  def testSingleEmptyChunk(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(chunked.Read(buf), b"")
    self.assertIsNone(chunked.Read(buf))

  def testMultipleEmptyChunks(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"")
    chunked.Write(buf, b"")
    chunked.Write(buf, b"")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(chunked.Read(buf), b"")
    self.assertEqual(chunked.Read(buf), b"")
    self.assertEqual(chunked.Read(buf), b"")
    self.assertIsNone(chunked.Read(buf))


class ReadTest(absltest.TestCase):

  def testEmptyBuffer(self):
    buf = io.BytesIO()
    self.assertIsNone(chunked.Read(buf))

  def testIncorrectSizeTag(self):
    buf = io.BytesIO(b"\x00\xff\xee")
    with self.assertRaises(chunked.IncorrectSizeTagError):
      chunked.Read(buf)

  def testMalformedInpuWithMaxChunkSizeSet(self):
    buf = io.BytesIO(b"\xff" * 1024)

    with self.assertRaises(chunked.ChunkSizeTooBigError):
      chunked.Read(buf, max_chunk_size=1024)

  def testMalformedInputWithoutMaxChunkSizeSet(self):
    buf1 = io.BytesIO()
    chunked.Write(buf1, b"foobarbaz")

    buf2 = io.BytesIO(buf1.getvalue()[:-2])
    with self.assertRaises(chunked.ChunkTruncatedError):
      chunked.Read(buf2)


class ReadAllTest(absltest.TestCase):

  def testEmpty(self):
    buf = io.BytesIO()
    self.assertEmpty(list(chunked.ReadAll(buf)))

  def testSingle(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"foo")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(list(chunked.ReadAll(buf)), [b"foo"])

  def testMultiple(self):
    buf = io.BytesIO()
    chunked.Write(buf, b"foo")
    chunked.Write(buf, b"bar")
    chunked.Write(buf, b"quux")

    buf.seek(0, io.SEEK_SET)
    self.assertEqual(list(chunked.ReadAll(buf)), [b"foo", b"bar", b"quux"])


class EncodeTest(absltest.TestCase):

  def testEmpty(self):
    chunks = [b""]

    buf = io.BytesIO(b"".join(map(chunked.Encode, chunks)))
    self.assertEqual(chunked.Read(buf), b"")
    self.assertIsNone(chunked.Read(buf))

  def testSingle(self):
    chunks = [b"foo"]

    buf = io.BytesIO(b"".join(map(chunked.Encode, chunks)))
    self.assertEqual(chunked.Read(buf), b"foo")
    self.assertIsNone(chunked.Read(buf))

  def testMultiple(self):
    chunks = [b"foo", b"bar", b"baz"]

    buf = io.BytesIO(b"".join(map(chunked.Encode, chunks)))
    self.assertEqual(chunked.Read(buf), b"foo")
    self.assertEqual(chunked.Read(buf), b"bar")
    self.assertEqual(chunked.Read(buf), b"baz")
    self.assertIsNone(chunked.Read(buf))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
import io
import itertools

from absl.testing import absltest

from grr_response_core.lib.util import io as ioutil


class ChunkTest(absltest.TestCase):

  def testSizeZero(self):
    with self.assertRaises(ValueError):
      list(ioutil.Chunk(io.BytesIO(b"foobar"), size=0))

  def testSizeNegative(self):
    with self.assertRaises(ValueError):
      list(ioutil.Chunk(io.BytesIO(b"foobar"), size=-42))

  def testEmpty(self):
    chunks = list(ioutil.Chunk(io.BytesIO(b""), size=3))
    self.assertEmpty(chunks)

  def testSingleChunk(self):
    chunks = list(ioutil.Chunk(io.BytesIO(b"foobar"), size=6))
    self.assertEqual(chunks, [b"foobar"])

  def testManyEvenChunks(self):
    chunks = list(ioutil.Chunk(io.BytesIO(b"foobar"), size=2))
    self.assertEqual(chunks, [b"fo", b"ob", b"ar"])

  def testManyUnevenChunks(self):
    chunks = list(ioutil.Chunk(io.BytesIO(b"foobarbaz"), size=2))
    self.assertEqual(chunks, [b"fo", b"ob", b"ar", b"ba", b"z"])


class UnchunkedTest(absltest.TestCase):

  def testEmpty(self):
    stream = ioutil.Unchunk(iter([]))
    self.assertEqual(stream.read(), b"")

  def testSingleChunkReadWhole(self):
    stream = ioutil.Unchunk(iter([b"foobar"]))
    self.assertEqual(stream.read(), b"foobar")

  def testSingleChunkReadByByte(self):
    stream = ioutil.Unchunk(iter([b"foo"]))
    self.assertEqual(stream.read(1), b"f")
    self.assertEqual(stream.read(1), b"o")
    self.assertEqual(stream.read(1), b"o")
    self.assertEqual(stream.read(1), b"")

  def testMultiChunkReadByByte(self):
    stream = ioutil.Unchunk(iter([b"foo", b"bar", b"baz"]))
    self.assertEqual(stream.read(1), b"f")
    self.assertEqual(stream.read(1), b"o")
    self.assertEqual(stream.read(1), b"o")
    self.assertEqual(stream.read(1), b"b")
    self.assertEqual(stream.read(1), b"a")
    self.assertEqual(stream.read(1), b"r")
    self.assertEqual(stream.read(1), b"b")
    self.assertEqual(stream.read(1), b"a")
    self.assertEqual(stream.read(1), b"z")
    self.assertEqual(stream.read(1), b"")

  def testMultiChunkReadWhole(self):
    stream = ioutil.Unchunk(iter([b"foo", b"bar", b"baz"]))
    self.assertEqual(stream.read(), b"foobarbaz")

  def testMultiChunkReadLarge(self):
    stream = ioutil.Unchunk(itertools.cycle([b"\x42" * 1021]))
    self.assertEqual(stream.read(16777259), b"\x42" * 16777259)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""A module with utility functions for working with collections."""

from collections.abc import Callable, Iterable, Iterator
import itertools
from typing import TypeVar

T = TypeVar("T")
K = TypeVar("K")


def Flatten(iterator: Iterable[Iterable[T]]) -> Iterator[T]:
  """Flattens nested iterables into one iterator.

  Examples:
    >>> list(Flatten([[1, 2, 3], [4, 5, 6]]))
    [1, 2, 3, 4, 5, 6]

    >>> list([range(3), range(5), range(3)])
    [0, 1, 2, 0, 1, 2, 3, 4, 0, 1, 2]

  Args:
    iterator: An iterator of iterators to flatten.

  Yields:
    Items yielded by the given iterators.
  """
  for items in iterator:
    for item in items:
      yield item


def Trim(lst: list[T], limit: int) -> list[T]:
  """Trims a given list so that it is not longer than given limit.

  Args:
    lst: A list to trim.
    limit: A maximum number of elements in the list after trimming.

  Returns:
    A suffix of the input list that was trimmed.
  """
  limit = max(0, limit)

  clipping = lst[limit:]
  del lst[limit:]
  return clipping


def Group(items: Iterable[T], key: Callable[[T], K]) -> dict[K, T]:
  """Groups items by given key function.

  Args:
    items: An iterable or an iterator of items.
    key: A function which given each item will return the key.

  Returns:
    A dict with keys being each unique key and values being a list of items of
    that key.
  """
  result = {}

  for item in items:
    result.setdefault(key(item), []).append(item)

  return result


def Batch(items: Iterable[T], size: int) -> Iterator[list[T]]:
  """Divide items into batches of specified size.

  In case where number of items is not evenly divisible by the batch size, the
  last batch is going to be shorter.

  Args:
    items: An iterable or an iterator of items.
    size: A size of the returned batches.

  Yields:
    Lists of items with specified size.
  """
  batch = []

  for item in items:
    batch.append(item)
    if len(batch) == size:
      yield batch
      batch = []

  if batch:
    yield batch


def StartsWith(this: Iterable[T], that: Iterable[T]) -> bool:
  """Checks whether an items of one iterable are a prefix of another.

  Args:
    this: An iterable that needs to be checked.
    that: An iterable of which items must match the prefix of `this`.

  Returns:
    `True` if `that` is a prefix of `this`, `False` otherwise.
  """
  this_iter = iter(this)
  that_iter = iter(that)

  while True:
    try:
      this_value = next(that_iter)
    except StopIteration:
      return True

    try:
      that_value = next(this_iter)
    except StopIteration:
      return False

    if this_value != that_value:
      return False


def Unzip(iterable: Iterable[tuple[K, T]]) -> tuple[Iterable[K], Iterable[T]]:
  """Unzips specified iterable of pairs to pair of two iterables.

  This function is an inversion of the standard `zip` function and the following
  hold:

    * ∀ l, r. l, r == unzip(zip(l, r))
    * ∀ p. p == zip(unzip(p))

  Examples:
    >>> Unzip([("foo", 1), ("bar", 2), ("baz", 3)])
    (["foo", "bar", "baz"], [1, 2, 3])

  Args:
    iterable: An iterable of pairs to unzip.

  Returns:
    A pair of iterables after unzipping.
  """
  lefts = []
  rights = []

  for left, right in iterable:
    lefts.append(left)
    rights.append(right)

  return lefts, rights


def DictProduct(dictionary: dict[K, Iterable[T]]) -> Iterator[dict[K, T]]:
  """Computes a cartesian product of dict with iterable values.

  This utility function, accepts a dictionary with iterable values, computes
  cartesian products of these values and yields dictionaries of expanded values.

  Examples:
    >>> list(DictProduct({"a": [1, 2], "b": [3, 4]}))
    [{"a": 1, "b": 3}, {"a": 1, "b": 4}, {"a": 2, "b": 3}, {"a": 2, "b": 4}]

  Args:
    dictionary: A dictionary with iterable values.

  Yields:
    Dictionaries with values being a result of cartesian product of values of
    the input dictionary.
  """
  keys, values = Unzip(dictionary.items())
  for product_values in itertools.product(*values):
    yield dict(zip(keys, product_values))



#!/usr/bin/env python
"""Plist related rdfvalues."""

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import sysinfo_pb2


class PlistBoolDictEntry(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.PlistBoolDictEntry


class PlistStringDictEntry(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.PlistStringDictEntry


class LaunchdStartCalendarIntervalEntry(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.LaunchdStartCalendarIntervalEntry


class LaunchdKeepAlive(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.LaunchdKeepAlive
  rdf_deps = [
      PlistBoolDictEntry,
  ]


class LaunchdPlist(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.LaunchdPlist
  rdf_deps = [
      LaunchdKeepAlive,
      LaunchdStartCalendarIntervalEntry,
      PlistStringDictEntry,
  ]



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from google.protobuf import timestamp_pb2
from grr_response_core.lib.rdfvalues import wkt as rdf_wkt


def ToProtoTimestamp(rdf: rdf_wkt.Timestamp) -> timestamp_pb2.Timestamp:
  return rdf.AsPrimitiveProto()


def ToRDFTimestamp(proto: timestamp_pb2.Timestamp) -> rdf_wkt.Timestamp:
  return rdf_wkt.Timestamp.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""The various ReadLowLevel rdfvalues."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import read_low_level_pb2


class ReadLowLevelArgs(rdf_structs.RDFProtoStruct):
  """Arguments for ReadLowLevel flow."""

  protobuf = read_low_level_pb2.ReadLowLevelArgs
  rdf_deps = [
      rdfvalue.ByteSize,
  ]

  # MAX_RAW_DATA_BYTES sets the limit for requesting raw data for the client.
  # This limit amounts to 10 GiB. If the user needs more than that, they will
  # need to schedule more than one read and concatenate the data themselves.
  MAX_RAW_DATA_BYTES = 10 * 1024 * 1024 * 1024  # 10 GiB

  def Validate(self):
    if not self.HasField("path"):
      raise ValueError("No path provided")

    if not self.HasField("length"):
      raise ValueError("No length provided")

    if self.length <= 0:
      raise ValueError(f"Negative length ({self.length})")

    if self.length > self.MAX_RAW_DATA_BYTES:
      raise ValueError(
          f"Cannot read more than {self.MAX_RAW_DATA_BYTES} bytes "
          f"({self.length} bytes requested"
      )


class ReadLowLevelFlowResult(rdf_structs.RDFProtoStruct):
  """Result returned by ReadLowLevel."""

  protobuf = read_low_level_pb2.ReadLowLevelFlowResult
  rdf_deps = []


class ReadLowLevelRequest(rdf_structs.RDFProtoStruct):
  """Request for ReadLowLevel action."""

  protobuf = read_low_level_pb2.ReadLowLevelRequest
  rdf_deps = [
      rdfvalue.ByteSize,
  ]


class ReadLowLevelResult(rdf_structs.RDFProtoStruct):
  """Result for ReadLowLevel action."""

  protobuf = read_low_level_pb2.ReadLowLevelResult
  rdf_deps = [
      rdf_client.BufferReference,
      rdfvalue.HashDigest,
  ]



#!/usr/bin/env python
"""Helper library for config testing."""

import copy
import logging
from unittest import mock

from grr_response_core import config
from grr_response_core.lib import config_lib
from grr_response_core.lib import utils
from grr.test_lib import test_lib


class BuildConfigTestsBase(test_lib.GRRBaseTest):
  """Base for config functionality tests."""

  exceptions = [
      # Those keys are maybe passphrase protected so we need to skip.
      "PrivateKeys.executable_signing_private_key",
  ]

  # For all the resource filters to work you need the grr-response-templates
  # package, which is big.
  disabled_filters = ["resource", "file"]

  def ValidateConfig(self, config_file=None):
    """Iterate over all the sections in the config file and validate them."""
    logging.debug("Processing %s", config_file)

    if isinstance(config_file, config_lib.GrrConfigManager):
      conf_obj = config_file
    else:
      conf_obj = config.CONFIG.MakeNewConfig()
      conf_obj.Initialize(filename=config_file, reset=True)

    with utils.MultiStubber(
        (config, "CONFIG", conf_obj), (config_lib, "_CONFIG", conf_obj)
    ):
      all_sections = conf_obj.GetSections()
      errors = conf_obj.Validate(sections=all_sections)

    return errors

  def ValidateConfigs(self, configs):
    test_filter_map = copy.deepcopy(config_lib.ConfigFilter.classes_by_name)
    for filter_name in self.disabled_filters:
      test_filter_map[filter_name] = config_lib.ConfigFilter

    with mock.patch.object(
        config_lib.ConfigFilter, "classes_by_name", test_filter_map
    ):
      for config_file in configs:
        errors = self.ValidateConfig(config_file)

        for exception in self.exceptions:
          errors.pop(exception, None)

        if errors:
          logging.info("Validation of %s returned errors:", config_file)
          for config_entry, error in errors.items():
            logging.info("%s:", config_entry)
            logging.info("%s", error)

          self.fail(
              "Validation of %s returned errors: %s" % (config_file, errors)
          )



#!/usr/bin/env python
"""Common tests for stats-collector implementations."""

import abc
import time
from unittest import mock

from absl.testing import absltest

from grr_response_core.lib.rdfvalues import stats as rdf_stats
from grr_response_core.stats import metrics
from grr.test_lib import stats_test_lib

_INF = float("inf")


class StatsCollectorTest(
    stats_test_lib.StatsCollectorTestMixin,
    absltest.TestCase,
    metaclass=abc.ABCMeta,
):
  """Stats collection tests.

  Each test method has uniquely-named metrics to accommodate implementations
  that do not support re-definition of metrics.

  For Events, the exact boundaries of Distribution bins are not tested. For
  these histogram metrics, it is acceptable that different implementations have
  slightly different behavior, e.g. one uses lower or equal while another uses
  strictly lower for bounds of bins. This allows integration with third-party
  metric libraries.
  """

  def setUp(self):
    super().setUp()

    self._mock_time = 100.0
    time_patcher = mock.patch.object(time, "time", lambda: self._mock_time)
    time_patcher.start()
    self.addCleanup(time_patcher.stop)

  @abc.abstractmethod
  def _CreateStatsCollector(self):
    """Creates a new stats collector."""

  def _Sleep(self, n):
    """Simulates sleeping for a given number of seconds."""
    self._mock_time += n

  def testSimpleCounter(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testSimpleCounter_counter")

    self.assertEqual(0, counter.GetValue())

    for _ in range(5):
      counter.Increment()
    self.assertEqual(5, counter.GetValue())

    counter.Increment(2)
    self.assertEqual(7, counter.GetValue())

  def testDecrementingCounterRaises(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testDecrementingCounterRaises_counter")

    with self.assertRaises(ValueError):
      counter.Increment(-1)

  def testCounterWithFields(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter(
          "testCounterWithFields_counter", fields=[("dimension", str)]
      )

    # Test that default values for any fields values are 0."
    self.assertEqual(0, counter.GetValue(fields=["a"]))
    self.assertEqual(0, counter.GetValue(fields=["b"]))

    for _ in range(5):
      counter.Increment(fields=["dimension_value_1"])
    self.assertEqual(5, counter.GetValue(fields=["dimension_value_1"]))

    counter.Increment(2, fields=["dimension_value_1"])
    self.assertEqual(7, counter.GetValue(fields=["dimension_value_1"]))

    counter.Increment(2, fields=["dimension_value_2"])
    self.assertEqual(2, counter.GetValue(fields=["dimension_value_2"]))
    # Check that previously set values with other fields are not affected.
    self.assertEqual(7, counter.GetValue(fields=["dimension_value_1"]))

  def testCounterWithBoolFields(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter(
          f"{self.testCounterWithBoolFields.__name__}_COUNTER",
          fields=[("is_foo", bool)],
      )

    self.assertEqual(counter.GetValue([False]), 0)
    self.assertEqual(counter.GetValue([True]), 0)

    counter.Increment(fields=[True])

    self.assertEqual(counter.GetValue([False]), 0)
    self.assertEqual(counter.GetValue([True]), 1)

    counter.Increment(fields=[False])

    self.assertEqual(counter.GetValue([False]), 1)
    self.assertEqual(counter.GetValue([True]), 1)

    counter.Increment(fields=[False])

    self.assertEqual(counter.GetValue([False]), 2)
    self.assertEqual(counter.GetValue([True]), 1)

  def testSimpleGauge(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      int_gauge = metrics.Gauge("testSimpleGauge_int_gauge", int)
      float_gauge = metrics.Gauge("testSimpleGauge_float_gauge", float)

    self.assertEqual(0, int_gauge.GetValue())
    self.assertEqual(0.0, float_gauge.GetValue())
    int_gauge.SetValue(42)
    float_gauge.SetValue(42.3)

    self.assertEqual(42, int_gauge.GetValue())
    self.assertAlmostEqual(42.3, float_gauge.GetValue())

    # At least default Python type checking is enforced in gauges:
    # we can't assign string to int
    with self.assertRaises(ValueError):
      int_gauge.SetValue("some")

  def testGaugeWithFields(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      int_gauge = metrics.Gauge(
          "testGaugeWithFields_int_gauge", int, fields=[("dimension", str)]
      )

    self.assertEqual(0, int_gauge.GetValue(fields=["dimension_value_1"]))
    self.assertEqual(0, int_gauge.GetValue(fields=["dimesnioN_value_2"]))

    int_gauge.SetValue(1, fields=["dimension_value_1"])
    int_gauge.SetValue(2, fields=["dimension_value_2"])

    self.assertEqual(1, int_gauge.GetValue(fields=["dimension_value_1"]))
    self.assertEqual(2, int_gauge.GetValue(fields=["dimension_value_2"]))

  def testGaugeWithCallback(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      int_gauge = metrics.Gauge("testGaugeWithCallback_int_gauge", int)
      float_gauge = metrics.Gauge("testGaugeWithCallback_float_gauge", float)

    self.assertEqual(0, int_gauge.GetValue())
    self.assertEqual(0.0, float_gauge.GetValue())

    int_gauge.SetCallback(lambda: 42)
    float_gauge.SetCallback(lambda: 42.3)

    self.assertEqual(42, int_gauge.GetValue())
    self.assertAlmostEqual(42.3, float_gauge.GetValue())

  def testSimpleEventMetric(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      event_metric = metrics.Event(
          "testSimpleEventMetric_event_metric", bins=[0.0, 0.1, 0.2]
      )

    data = event_metric.GetValue()
    self.assertAlmostEqual(0, data.sum)
    self.assertEqual(0, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 0, 0.2: 0}, data.bins_heights)

    event_metric.RecordEvent(0.15)
    data = event_metric.GetValue()
    self.assertAlmostEqual(0.15, data.sum)
    self.assertEqual(1, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 1, 0.2: 0}, data.bins_heights)

    event_metric.RecordEvent(0.5)
    data = event_metric.GetValue()
    self.assertAlmostEqual(0.65, data.sum)
    self.assertEqual(2, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 1, 0.2: 1}, data.bins_heights)

    event_metric.RecordEvent(-0.1)
    data = event_metric.GetValue()
    self.assertAlmostEqual(0.55, data.sum)
    self.assertEqual(3, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 1, 0.0: 0, 0.1: 1, 0.2: 1}, data.bins_heights)

  def testEventMetricWithFields(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      event_metric = metrics.Event(
          "testEventMetricWithFields_event_metric",
          bins=[0.0, 0.1, 0.2],
          fields=[("dimension", str)],
      )

    data = event_metric.GetValue(fields=["dimension_value_1"])
    self.assertAlmostEqual(0, data.sum)
    self.assertEqual(0, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 0, 0.2: 0}, data.bins_heights)

    event_metric.RecordEvent(0.15, fields=["dimension_value_1"])
    event_metric.RecordEvent(0.25, fields=["dimension_value_2"])

    data = event_metric.GetValue(fields=["dimension_value_1"])
    self.assertAlmostEqual(0.15, data.sum)
    self.assertEqual(1, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 1, 0.2: 0}, data.bins_heights)

    data = event_metric.GetValue(fields=["dimension_value_2"])
    self.assertAlmostEqual(0.25, data.sum)
    self.assertEqual(1, data.count)
    self.assertEqual([-_INF, 0.0, 0.1, 0.2], list(data.bins))
    self.assertEqual({-_INF: 0, 0.0: 0, 0.1: 0, 0.2: 1}, data.bins_heights)

  def testRaisesOnImproperFieldsUsage1(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testRaisesOnImproperFieldsUsage1_counter")
      int_gauge = metrics.Gauge(
          "testRaisesOnImproperFieldsUsage1_int_gauge", int
      )
      event_metric = metrics.Event(
          "testRaisesOnImproperFieldsUsage1_event_metric"
      )

    # Check for counters
    with self.assertRaises(ValueError):
      counter.GetValue(fields=["a"])

    # Check for gauges
    with self.assertRaises(ValueError):
      int_gauge.GetValue(fields=["a"])

    # Check for event metrics
    with self.assertRaises(ValueError):
      event_metric.GetValue(fields=["a", "b"])

  def testRaisesOnImproperFieldsUsage2(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter(
          "testRaisesOnImproperFieldsUsage2_counter",
          fields=[("dimension", str)],
      )
      int_gauge = metrics.Gauge(
          "testRaisesOnImproperFieldsUsage2_int_gauge",
          int,
          fields=[("dimension", str)],
      )
      event_metric = metrics.Event(
          "testRaisesOnImproperFieldsUsage2_event_metric",
          fields=[("dimension", str)],
      )

    # Check for counters
    with self.assertRaises(ValueError):
      counter.GetValue()
    with self.assertRaises(ValueError):
      counter.GetValue(fields=["a", "b"])

    # Check for gauges
    with self.assertRaises(ValueError):
      int_gauge.GetValue()
    with self.assertRaises(ValueError):
      int_gauge.GetValue(fields=["a", "b"])

    # Check for event metrics
    with self.assertRaises(ValueError):
      event_metric.GetValue()
    with self.assertRaises(ValueError):
      event_metric.GetValue(fields=["a", "b"])

  def testGetAllMetricsMetadataWorksCorrectlyOnSimpleMetrics(self):
    counter_name = "testGAMM_SimpleMetrics_counter"
    int_gauge_name = "testGAMM_SimpleMetrics_int_gauge"
    event_metric_name = "testGAMM_SimpleMetrics_event_metric"

    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      metrics.Counter(counter_name)
      metrics.Gauge(int_gauge_name, int, fields=[("dimension", str)])
      metrics.Event(event_metric_name)

    metadatas = self.collector.GetAllMetricsMetadata()
    self.assertEqual(
        metadatas[counter_name].metric_type,
        rdf_stats.MetricMetadata.MetricType.COUNTER,
    )
    self.assertFalse(metadatas[counter_name].fields_defs)

    self.assertEqual(
        metadatas[int_gauge_name].metric_type,
        rdf_stats.MetricMetadata.MetricType.GAUGE,
    )
    self.assertEqual(
        metadatas[int_gauge_name].fields_defs,
        [
            rdf_stats.MetricFieldDefinition(
                field_name="dimension",
                field_type=rdf_stats.MetricFieldDefinition.FieldType.STR,
            )
        ],
    )

    self.assertEqual(
        metadatas[event_metric_name].metric_type,
        rdf_stats.MetricMetadata.MetricType.EVENT,
    )
    self.assertFalse(metadatas[event_metric_name].fields_defs)

  def testGetMetricFieldsWorksCorrectly(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter(
          "testGetMetricFieldsWorksCorrectly_counter",
          fields=[("dimension1", str), ("dimension2", str)],
      )
      int_gauge = metrics.Gauge(
          "testGetMetricFieldsWorksCorrectly_int_gauge",
          int,
          fields=[("dimension", str)],
      )
      event_metric = metrics.Event(
          "testGetMetricFieldsWorksCorrectly_event_metric",
          fields=[("dimension", str)],
      )

    counter.Increment(fields=["b", "b"])
    counter.Increment(fields=["a", "c"])
    self.assertCountEqual([("a", "c"), ("b", "b")], counter.GetFields())

    int_gauge.SetValue(20, fields=["a"])
    int_gauge.SetValue(30, fields=["b"])
    self.assertCountEqual([("a",), ("b",)], int_gauge.GetFields())

    event_metric.RecordEvent(0.1, fields=["a"])
    event_metric.RecordEvent(0.1, fields=["b"])
    self.assertCountEqual([("a",), ("b",)], event_metric.GetFields())

  def testCountingDecorator(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testCountingDecorator_counter")

    @counter.Counted()
    def CountedFunc():
      pass

    for _ in range(10):
      CountedFunc()

    self.assertEqual(counter.GetValue(), 10)

  def testSuccessesCountingDecorator(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testCountingDecorator_successes_counter")

    @counter.SuccessesCounted()
    def CountedFunc(should_raise):
      if should_raise:
        raise RuntimeError("foo")

    for i in range(10):
      if i % 2 == 0:
        with self.assertRaises(RuntimeError):
          CountedFunc(True)
      else:
        CountedFunc(False)

    # Failing calls shouldn't increment the counter.
    self.assertEqual(counter.GetValue(), 5)

  def testErrorsCountingDecorator(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testCountingDecorator_errors_counter")

    @counter.SuccessesCounted()
    def CountedFunc(should_raise):
      if should_raise:
        raise RuntimeError("foo")

    for i in range(10):
      if i % 2 == 0:
        with self.assertRaises(RuntimeError):
          CountedFunc(True)
      else:
        CountedFunc(False)

    # Non-failing calls shouldn't increment the counter.
    self.assertEqual(counter.GetValue(), 5)

  def testBinnedTimings(self):
    event_metric_name = "testMaps_event_metric"

    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      event_metric = metrics.Event(event_metric_name, bins=[0, 0.1, 0.2])

    @event_metric.Timed()
    def TimedFunc(n):
      self._Sleep(n)

    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 0, 0.1: 0, 0.2: 0})

    for _ in range(3):
      TimedFunc(0.01)

    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 3, 0.1: 0, 0.2: 0})

    TimedFunc(0.11)
    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 3, 0.1: 1, 0.2: 0})

  def testCombiningDecorators(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testCombiningDecorators_counter")
      event_metric = metrics.Event(
          "testCombiningDecorators_event_metric", bins=[0.0, 0.1, 0.2]
      )

    @event_metric.Timed()
    @counter.Counted()
    def OverdecoratedFunc(n):
      self._Sleep(n)

    OverdecoratedFunc(0.02)

    # Check if all vars get updated
    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 1, 0.1: 0, 0.2: 0})

    self.assertEqual(counter.GetValue(), 1)

  def testExceptionHandling(self):
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testExceptionHandling_counter")
      event_metric = metrics.Event(
          "testExceptionHandling_event_metric", bins=[0, 0.1, 0.2]
      )

    @event_metric.Timed()
    @counter.Counted()
    def RaiseFunc(n):
      self._Sleep(n)
      raise Exception()

    with self.assertRaises(Exception):
      RaiseFunc(0.11)

    # Check if all vars get updated
    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 0, 0.1: 1, 0.2: 0})

    self.assertEqual(counter.GetValue(), 1)

  def testMultipleFuncs(self):
    """Tests if multiple decorators produce aggregate stats."""
    with self.SetUpStatsCollector(self._CreateStatsCollector()):
      counter = metrics.Counter("testMultipleFuncs_counter")
      event_metric = metrics.Event(
          "testMultipleFuncs_event_metric", bins=[0, 1, 2]
      )

    @counter.Counted()
    def Func1(n):
      self._Sleep(n)

    @counter.Counted()
    def Func2(n):
      self._Sleep(n)

    @event_metric.Timed()
    def Func3(n):
      self._Sleep(n)

    @event_metric.Timed()
    def Func4(n):
      self._Sleep(n)

    Func1(0.1)
    Func2(0.1)
    self.assertEqual(counter.GetValue(), 2)

    Func3(0.1)
    Func4(1.1)
    m = event_metric.GetValue()
    self.assertEqual(m.bins_heights, {-_INF: 0, 0: 1, 1: 1, 2: 0})



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import webhistory as rdf_webhistory
from grr_response_proto import sysinfo_pb2


def ToProtoBrowserHistoryItem(
    rdf: rdf_webhistory.BrowserHistoryItem,
) -> sysinfo_pb2.BrowserHistoryItem:
  return rdf.AsPrimitiveProto()


def ToRDFBrowserHistoryItem(
    proto: sysinfo_pb2.BrowserHistoryItem,
) -> rdf_webhistory.BrowserHistoryItem:
  return rdf_webhistory.BrowserHistoryItem.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""This directory contains local site-specific implementations."""

# Local site-specific implementations that have to be imported to be registered
# should be imported in plugins.py.



#!/usr/bin/env python
"""Configuration parameters for the client."""


from grr_response_core.lib import config_lib
from grr_response_core.lib.rdfvalues import crypto as rdf_crypto

# General Client options.
config_lib.DEFINE_string(
    "Client.name", "GRR", "The name of the client. This will be used as a base "
    "name to generate many other default parameters such "
    "as binary names and service names. Note that on "
    "Linux we lowercase the name to confirm with most "
    "linux naming conventions.")

config_lib.DEFINE_string("Client.binary_name", "%(Client.name)",
                         "The name of the client binary.")

config_lib.DEFINE_string("Client.company_name", "GRR Project",
                         "The name of the company which made the client.")

config_lib.DEFINE_string("Client.description", "%(name) %(platform) %(arch)",
                         "A description of this specific client build.")

config_lib.DEFINE_string("Client.platform", "windows",
                         "The platform we are running on.")

config_lib.DEFINE_string("Client.arch", "amd64",
                         "The architecture we are running on.")

config_lib.DEFINE_string("Client.build_time", "Unknown",
                         "The time the client was built.")

config_lib.DEFINE_string("Client.deploy_time", "Unknown",
                         "The time the client was deployed.")

config_lib.DEFINE_string(
    "Client.build_environment", None,
    "The output of Uname.FromCurrentSystem.signature() "
    "on the system the client was built on.")

config_lib.DEFINE_integer("Client.rsa_key_length", 2048,
                          "The key length of the client keys in bits.")

config_lib.DEFINE_string(
    name="Client.install_path",
    default=r"%(SystemRoot|env)\\System32\\%(name)\\%(Template.version_string)",
    help="Where the client binaries are installed.")

config_lib.DEFINE_integer("Client.http_timeout", 100,
                          "Timeout for HTTP requests.")

config_lib.DEFINE_string("Client.plist_path",
                         "/Library/LaunchDaemons/%(Client.plist_filename)",
                         "Location of our launchctl plist.")

config_lib.DEFINE_string("Client.plist_filename", "%(Client.plist_label).plist",
                         "Filename of launchctl plist.")

config_lib.DEFINE_string(
    "Client.plist_label",
    "%(Client.plist_label_prefix).google.code.%(Client.name)",
    "Identifier label for launchd")

config_lib.DEFINE_string("Client.plist_label_prefix", "com",
                         "Domain for launchd label.")

config_lib.DEFINE_list(
    name="Client.proxy_servers",
    help="List of valid proxy servers the client should try.",
    default=[])

config_lib.DEFINE_integer("Client.max_post_size", 40000000,
                          "Maximum size of the post.")

config_lib.DEFINE_integer("Client.max_out_queue", 51200000,
                          "Maximum size of the output queue.")

config_lib.DEFINE_integer(
    "Client.foreman_check_frequency", 1800,
    "The minimum number of seconds before checking with "
    "the foreman for new work.")

config_lib.DEFINE_string(
    name="Client.tempfile_prefix",
    help="Prefix to use for temp files created by the GRR client.",
    default="tmp%(Client.name)")

config_lib.DEFINE_list(
    name="Client.tempdir_roots",
    help="List of temporary directories to use on the client.",
    default=["/var/tmp/"])

config_lib.DEFINE_string(
    name="Client.grr_tempdir",
    help="Default subdirectory in the temp directory to use for GRR.",
    default="%(Client.name)")

config_lib.DEFINE_list(
    name="Client.vfs_virtualroots",
    help=("If this is set for a VFS type, client VFS operations will always be"
          " relative to the given root. Format is os:/mount/disk."),
    default=[])

config_lib.DEFINE_bool(
    name="Client.use_filesystem_sandboxing",
    help="Whether to use the sandboxed implementation for filesystem parsing.",
    default=False)

config_lib.DEFINE_bool(
    name="Client.use_memory_sandboxing",
    help="Whether to use the sandboxed implementation for memory scanning.",
    default=False)

config_lib.DEFINE_string(
    name="Client.unprivileged_user",
    help="Name of (UNIX) user to run sandboxed code as.",
    default="")

config_lib.DEFINE_string(
    name="Client.unprivileged_group",
    help="Name of (UNIX) group to run sandboxed code as.",
    default="")

# Windows client specific options.
config_lib.DEFINE_string(
    "Client.config_hive",
    "HKEY_LOCAL_MACHINE",
    help="The registry hive where the client "
    "configuration will be stored.")

config_lib.DEFINE_string(
    "Client.config_key",
    r"Software\\%(Client.name)",
    help="The registry key where  client configuration "
    "will be stored.")

config_lib.DEFINE_semantic_value(
    rdf_crypto.RSAPublicKey,
    "Client.executable_signing_public_key",
    help="public key for verifying executable signing.")

config_lib.DEFINE_semantic_value(
    rdf_crypto.RSAPrivateKey,
    "PrivateKeys.executable_signing_private_key",
    help="Private keys for signing executables. NOTE: This "
    "key is usually kept offline and is thus not present in the "
    "configuration file.")

config_lib.DEFINE_integer(
    "Client.gc_frequency", 10,
    "Defines how often the client calls garbage collection (seconds).")

config_lib.DEFINE_string(
    "Client.transaction_log_file",
    "%(Logging.path)/transaction.log",
    "The file where we write the agent transaction log.",
)

config_lib.DEFINE_integer(
    "Network.api", 3, "The version of the network protocol the client "
    "uses.")

# Installer options.
config_lib.DEFINE_string(
    name="Installer.logfile",
    default="%(Logging.path)/%(Client.name)_installer.txt",
    help=("A specific log file which is used for logging the "
          "installation process."))

config_lib.DEFINE_string(
    "Client.fleetspeak_unsigned_services_regkey",
    "HKEY_LOCAL_MACHINE\\Software\\FleetspeakClient\\textservices",
    "Registry key (on Windows) where Fleetspeak expects services "
    "to write their unsigned configs to.")

config_lib.DEFINE_string(
    "Client.fleetspeak_unsigned_config_fname",
    "%(Client.name)_fleetspeak_service_config.txt",
    "File-name for the Fleetspeak service config generated "
    "when repacking templates.")

config_lib.DEFINE_list(
    "Client.allowed_commands",
    default=[],
    help="Commands that the client is allowed to execute. Each command must be "
    "specified in the format that is supported by the Python's `shlex` module.")

config_lib.DEFINE_string(
    "Client.interrogate_trigger_path",
    default="",
    help="When set, the client will check for the presence of this file at "
    "every startup and, if the file is present, will remove it and trigger "
    "an interrogate on the server.")

# osquery options.

config_lib.DEFINE_string(
    "Osquery.path", default="", help="A path to the osquery executable.")

config_lib.DEFINE_integer(
    "Osquery.max_chunk_size",
    default=1024 * 1024,  # 1 MiB.
    help=("A size (in bytes) of maximum response size. Queries for which the "
          "output exceedes the specified limit are going to be divided into "
          "multiple responses."))



#!/usr/bin/env python
"""Stats-related client rdfvalues."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import jobs_pb2


class CpuSeconds(rdf_structs.RDFProtoStruct):
  """CPU usage is reported as both a system and user components."""

  protobuf = jobs_pb2.CpuSeconds


class ClientResources(rdf_structs.RDFProtoStruct):
  """An RDFValue class representing the client resource usage."""

  protobuf = jobs_pb2.ClientResources
  rdf_deps = [
      rdf_client.ClientURN,
      CpuSeconds,
      rdfvalue.SessionID,
  ]



#!/usr/bin/env python
"""A module with assertion functions for checking preconditions."""

import collections
from collections import abc
from collections.abc import Sized
import re


def AssertType(value, expected_type):
  """Ensures that given value has certain type.

  Args:
    value: A value to assert the type for.
    expected_type: An expected type for the given value.

  Raises:
    TypeError: If given value does not have the expected type.
  """
  if not isinstance(value, expected_type):
    message = "Expected type `%r`, but got value `%r` of type `%s`"
    message %= (expected_type, value, type(value))
    raise TypeError(message)


def AssertOptionalType(value, expected_type):
  """Ensures that given value, if not `None`, has certain type.

  Args:
    value: A value or `None` to assert the type for.
    expected_type: An expected type for the given value.

  Raises:
    TypeError: If given value is not `None` and does not have the expected type.
  """
  if value is not None:
    AssertType(value, expected_type)


def AssertIterableType(iterable, expected_item_type):
  """Ensures that given iterable container has certain type.

  Args:
    iterable: An iterable container to assert the type for.
    expected_item_type: An expected type of the container items.

  Raises:
    TypeError: If given container does is not an iterable or its items do not
               have the expected type.
  """
  # We do not consider iterators to be iterables even though Python does. An
  # "iterable" should be a type that can be iterated (that is: an iterator can
  # be constructed for them). Iterators should not be considered to be iterable
  # because it makes no sense to construct an iterator for iterator. The most
  # important practical implication is that act of iterating an iterator drains
  # it whereas act of iterating the iterable does not.
  if isinstance(iterable, collections.abc.Iterator):
    message = "Expected iterable container but got iterator `%s` instead"
    message %= iterable
    raise TypeError(message)

  AssertType(iterable, abc.Iterable)
  for item in iterable:
    AssertType(item, expected_item_type)


def AssertDictType(dct, expected_key_type, expected_value_type):
  """Ensures that given dictionary is actually a dictionary of specified type.

  Args:
    dct: A dictionary to assert the type for.
    expected_key_type: An expected type for dictionary keys.
    expected_value_type: An expected type for dictionary values.

  Raises:
    TypeError: If given dictionary is not really a dictionary or not all its
               keys and values have the expected type.
  """
  AssertType(dct, abc.Mapping)
  for key, value in dct.items():
    AssertType(key, expected_key_type)
    AssertType(value, expected_value_type)


def AssertNotEmpty(typename, value):
  """Raises, if the given value is empty or has no __len__."""
  AssertType(value, Sized)
  if len(value) == 0:  # pylint: disable=g-explicit-length-test
    message = "Expected {} `{}` to be non-empty".format(typename, value)
    raise ValueError(message)


def _ValidateStringId(typename, value):
  AssertType(value, str)
  AssertNotEmpty(typename, value)


def ValidateClientId(client_id):
  """Raises, if the given value is not a valid ClientId string."""
  _ValidateStringId("client_id", client_id)
  # TODO(hanuszczak): Eventually, we should allow only either lower or upper
  # case letters in the client id.
  if re.match(r"^C\.[0-9a-fA-F]{16}$", client_id) is None:
    raise ValueError("Client id has incorrect format: `%s`" % client_id)


def ValidateFlowId(flow_id):
  """Raises, if the given value is not a valid FlowId string."""
  _ValidateStringId("flow_id", flow_id)
  if (
      len(flow_id) not in [8, 16]
      or re.match(r"^[0-9a-fA-F]*$", flow_id) is None
  ):
    raise ValueError("Flow id has incorrect format: `%s`" % flow_id)



#!/usr/bin/env python
"""Test client RDFValues."""

import platform
import socket
from unittest import mock

from absl import app
from absl.testing import absltest
import psutil

from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import client_fs as rdf_client_fs
from grr_response_core.lib.rdfvalues import client_network as rdf_client_network
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr_response_proto import knowledge_base_pb2
from grr.test_lib import test_lib


class UserTests(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  """Test the User ProtoStruct implementation."""

  rdfvalue_class = rdf_client.User

  def GenerateSample(self, number=0):
    result = rdf_client.User(username="user%s" % number)
    result.desktop = "User Desktop %s" % number

    return result

  def testKBUserBackwardsCompatibility(self):
    """Check User can be created from deprecated KBUser."""
    kbuser = rdf_client.KnowledgeBaseUser()
    kbuser.username = "user1"
    kbuser.desktop = "User Desktop 1"

    user = rdf_client.User(kbuser)

    self.assertEqual(user.username, "user1")
    self.assertEqual(user.desktop, "User Desktop 1")

  def testCompatibility(self):
    proto = knowledge_base_pb2.User(username="user1")
    proto.desktop = "User Desktop 1"

    serialized = proto.SerializeToString()

    rdf_from_serialized = rdf_client.User.FromSerializedBytes(serialized)

    self.assertEqual(rdf_from_serialized.username, proto.username)
    self.assertEqual(rdf_from_serialized.desktop, proto.desktop)

    rdf_direct = rdf_client.User(username="user1", desktop="User Desktop 1")

    self.assertEqual(rdf_from_serialized, rdf_direct)

  def testTimeEncoding(self):
    fast_proto = rdf_client.User(username="user")

    datetime = rdfvalue.RDFDatetime.FromHumanReadable("2013-04-05 16:00:03")

    # Check that we can coerce an int to an RDFDatetime.
    # TODO(hanuszczak): Yeah, but why would we...?
    fast_proto.last_logon = datetime.AsMicrosecondsSinceEpoch()

    self.assertEqual(fast_proto.last_logon, datetime)

    # Check that this is backwards compatible with the old protobuf library.
    proto = knowledge_base_pb2.User()
    proto.ParseFromString(fast_proto.SerializeToBytes())

    # Old implementation should just see the last_logon field as an integer.
    self.assertIsInstance(proto.last_logon, int)
    self.assertEqual(proto.last_logon, datetime.AsMicrosecondsSinceEpoch())

    # fast protobufs interoperate with old serialized formats.
    serialized_data = proto.SerializeToString()
    fast_proto = rdf_client.User.FromSerializedBytes(serialized_data)
    self.assertIsInstance(fast_proto.last_logon, rdfvalue.RDFDatetime)
    self.assertEqual(fast_proto.last_logon, datetime.AsMicrosecondsSinceEpoch())

  def testPrettyPrintMode(self):

    for mode, result in [
        (0o775, "-rwxrwxr-x"),
        (0o75, "----rwxr-x"),
        (0, "----------"),
        # DIR
        (0o40775, "drwxrwxr-x"),
        # SUID
        (35232, "-rwSr-----"),
        # GID
        (34208, "-rw-r-S---"),
        # CHR
        (9136, "crw-rw---T"),
        # BLK
        (25008, "brw-rw----"),
        # FIFO
        (4516, "prw-r--r--"),
        # Socket
        (49663, "srwxrwxrwx"),
        # Sticky
        (33791, "-rwxrwxrwt"),
        # Sticky, not x
        (33784, "-rwxrwx--T"),
    ]:
      value = rdf_client_fs.StatMode(mode)
      self.assertEqual(str(value), result)


class ClientURNTests(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  """Test the ClientURN."""

  rdfvalue_class = rdf_client.ClientURN

  def GenerateSample(self, number=0):
    return rdf_client.ClientURN("C.%016X" % number)

  def testInitialization(self):
    """ClientURNs don't allow empty init so we override the default test."""

    self.rdfvalue_class("C.00aaeccbb45f33a3")

    # Initialize from another instance.
    sample = self.GenerateSample()

    self.CheckRDFValue(self.rdfvalue_class(sample), sample)

  def testURNValidation(self):
    # These should all come out the same: C.00aaeccbb45f33a3
    test_set = [
        "C.00aaeccbb45f33a3",
        "C.00aaeccbb45f33a3".upper(),
        "c.00aaeccbb45f33a3",
        "C.00aaeccbb45f33a3 ",
    ]
    results = []
    for urnstr in test_set:
      results.append(rdf_client.ClientURN(urnstr))
      results.append(rdf_client.ClientURN("aff4:/%s" % urnstr))

    self.assertLen(results, len(test_set) * 2)

    # Check all are identical
    self.assertTrue(all([x == results[0] for x in results]))

    # Check we can handle URN as well as string
    rdf_client.ClientURN(rdf_client.ClientURN(test_set[0]))

    error_set = [
        "B.00aaeccbb45f33a3",
        "c.00accbb45f33a3",
        "aff5:/C.00aaeccbb45f33a3",
    ]

    for badurn in error_set:
      self.assertRaises(type_info.TypeValueError, rdf_client.ClientURN, badurn)


class NetworkAddressTests(
    rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest
):
  """Test the NetworkAddress."""

  rdfvalue_class = rdf_client_network.NetworkAddress

  def GenerateSample(self, number=0):
    return rdf_client_network.NetworkAddress(
        human_readable_address="192.168.0.%s" % number
    )

  def testIPv4(self):
    sample = rdf_client_network.NetworkAddress(
        human_readable_address="192.168.0.1"
    )
    self.assertEqual(
        sample.address_type, rdf_client_network.NetworkAddress.Family.INET
    )
    # Equal to socket.inet_pton(socket.AF_INET, "192.168.0.1"), which is
    # unavailable on Windows.
    self.assertEqual(sample.packed_bytes, b"\xc0\xa8\x00\x01")

    self.assertEqual(sample.human_readable_address, "192.168.0.1")

    self.CheckRDFValue(self.rdfvalue_class(sample), sample)

  def testIPv6(self):
    ipv6_addresses = ["fe80::202:b3ff:fe1e:8329", "::1"]
    # Equal to socket.inet_pton(socket.AF_INET6, address), which is unavailable
    # on Windows.
    expected_addresses = [
        b"\xfe\x80\x00\x00\x00\x00\x00\x00\x02\x02\xb3\xff\xfe\x1e\x83\x29",
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01",
    ]
    for address, expected in zip(ipv6_addresses, expected_addresses):
      sample = rdf_client_network.NetworkAddress(human_readable_address=address)
      self.assertEqual(
          sample.address_type, rdf_client_network.NetworkAddress.Family.INET6
      )
      self.assertEqual(sample.packed_bytes, expected)

      self.assertEqual(sample.human_readable_address, address)

      self.CheckRDFValue(self.rdfvalue_class(sample), sample)


class UnameTests(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  """Test the Uname."""

  rdfvalue_class = rdf_client.Uname

  def GenerateSample(self, number=0):
    # Make the hostname slightly different for comparison tests.
    result = self.rdfvalue_class.FromCurrentSystem()
    parts = result.fqdn.split(".")
    parts[0] += str(number)
    result.fqdn = ".".join(parts)
    return result

  def testSignature(self):
    sample = self.GenerateSample()
    self.assertEqual(sample.signature(), sample.pep425tag)

    # We do not support old protos without a signature.
    sample.pep425tag = None
    self.assertRaises(ValueError, sample.signature)

  def testGetFQDN(self):
    with mock.patch.object(socket, "getfqdn", return_value="foo.bar.baz"):
      uname = self.rdfvalue_class.FromCurrentSystem()
      self.assertEqual(uname.fqdn, "foo.bar.baz")

  def testGetFQDN_Localhost(self):
    with mock.patch.object(
        socket, "getfqdn", return_value=rdf_client._LOCALHOST
    ):
      with mock.patch.object(socket, "gethostname", return_value="foo"):
        uname = self.rdfvalue_class.FromCurrentSystem()
        self.assertEqual(uname.fqdn, "foo")


class ProcessTest(absltest.TestCase):

  def testFromPsutilProcess(self):

    p = psutil.Process()
    res = rdf_client.Process.FromPsutilProcess(p)

    int_fields = [
        "pid",
        "ppid",
        "ctime",
        "num_threads",
        "user_cpu_time",
        "system_cpu_time",
        "RSS_size",
        "VMS_size",
        "memory_percent",
    ]

    if platform.system() != "Windows":
      int_fields.extend([
          "real_uid",
          "effective_uid",
          "saved_uid",
          "real_gid",
          "effective_gid",
          "saved_gid",
      ])

    for field in int_fields:
      self.assertGreater(
          getattr(res, field),
          0,
          "rdf_client.Process.{} is not greater than 0, got {!r}.".format(
              field, getattr(res, field)
          ),
      )

    string_fields = ["name", "exe", "cmdline", "cwd", "username"]

    if platform.system() != "Windows":
      string_fields.append("terminal")

    for field in string_fields:
      self.assertNotEqual(
          getattr(res, field),
          "",
          "rdf_client.Process.{} is the empty string.".format(field),
      )

    # Prevent flaky tests by allowing "sleeping" as state of current process.
    self.assertIn(res.status, ["running", "sleeping"])


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
import io
import os

from absl.testing import absltest

from grr_response_core.lib.util import temp


class TempDirPathTest(absltest.TestCase):

  def testEmptyDirCreation(self):
    dirpath = temp.TempDirPath()

    self.assertTrue(os.path.exists(dirpath))
    self.assertEmpty(os.listdir(dirpath))

    os.rmdir(dirpath)

  def testPrefix(self):
    dirpath = temp.TempDirPath(prefix="foo")

    self.assertTrue(os.path.exists(dirpath))
    self.assertStartsWith(os.path.basename(dirpath), "foo")

    os.rmdir(dirpath)

  def testSuffix(self):
    dirpath = temp.TempDirPath(suffix="foo")

    self.assertTrue(os.path.exists(dirpath))
    self.assertEndsWith(os.path.basename(dirpath), "foo")

    os.rmdir(dirpath)

  def testPrefixAndSuffix(self):
    dirpath = temp.TempDirPath(prefix="foo", suffix="bar")

    self.assertTrue(os.path.exists(dirpath))
    self.assertStartsWith(os.path.basename(dirpath), "foo")
    self.assertEndsWith(os.path.basename(dirpath), "bar")

    os.rmdir(dirpath)


class TempFilePathTest(absltest.TestCase):

  def testEmptyFileCreation(self):
    filepath = temp.TempFilePath()

    self.assertTrue(os.path.exists(filepath))
    with io.open(filepath, "rb") as filedesc:
      self.assertEmpty(filedesc.read())

    os.remove(filepath)

  def testPrefix(self):
    filepath = temp.TempFilePath(prefix="foo")

    self.assertTrue(os.path.exists(filepath))
    self.assertStartsWith(os.path.basename(filepath), "foo")

    os.remove(filepath)

  def testSuffix(self):
    filepath = temp.TempFilePath(suffix="foo")

    self.assertTrue(os.path.exists(filepath))
    self.assertEndsWith(os.path.basename(filepath), "foo")

    os.remove(filepath)

  def testPrefixAndSuffix(self):
    filepath = temp.TempFilePath(prefix="foo", suffix="bar")

    self.assertTrue(os.path.exists(filepath))
    self.assertStartsWith(os.path.basename(filepath), "foo")
    self.assertEndsWith(os.path.basename(filepath), "bar")

    os.remove(filepath)

  def testDir(self):
    with temp.AutoTempDirPath() as dirpath:
      filepath = temp.TempFilePath(dir=dirpath)

      self.assertTrue(os.path.exists(filepath))
      self.assertStartsWith(filepath, dirpath)

      os.remove(filepath)


class AutoTempDirPathTest(absltest.TestCase):

  def testEmptyDirCreation(self):
    with temp.AutoTempDirPath() as dirpath:
      self.assertTrue(os.path.exists(dirpath))
      self.assertEmpty(os.listdir(dirpath))

  def testPrefix(self):
    with temp.AutoTempDirPath(prefix="foo") as dirpath:
      self.assertTrue(os.path.exists(dirpath))
      self.assertStartsWith(os.path.basename(dirpath), "foo")

  def testSuffix(self):
    with temp.AutoTempDirPath(suffix="foo") as dirpath:
      self.assertTrue(os.path.exists(dirpath))
      self.assertEndsWith(os.path.basename(dirpath), "foo")

  def testPrefixAndSuffix(self):
    with temp.AutoTempDirPath(prefix="foo", suffix="bar") as dirpath:
      self.assertTrue(os.path.exists(dirpath))
      self.assertStartsWith(os.path.basename(dirpath), "foo")
      self.assertEndsWith(os.path.basename(dirpath), "bar")

  def testRemovesEmptyDirs(self):
    with temp.AutoTempDirPath() as dirpath:
      self.assertTrue(os.path.exists(dirpath))

    self.assertFalse(os.path.exists(dirpath))

  def testRemovesNonEmptyDirs(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as dirpath:
      self.assertTrue(os.path.exists(dirpath))

      with io.open(os.path.join(dirpath, "foo"), "wb") as filedesc:
        filedesc.write(b"foo")

      with io.open(os.path.join(dirpath, "bar"), "wb") as filedesc:
        filedesc.write(b"bar")

    self.assertFalse(os.path.exists(dirpath))


class AutoTempFilePath(absltest.TestCase):

  def testEmptyFileCreation(self):
    with temp.AutoTempFilePath() as filepath:
      self.assertTrue(os.path.exists(filepath))
      with io.open(filepath, "rb") as filedesc:
        self.assertEmpty(filedesc.read())

  def testPrefix(self):
    with temp.AutoTempFilePath(prefix="foo") as filepath:
      self.assertTrue(os.path.exists(filepath))
      self.assertStartsWith(os.path.basename(filepath), "foo")

  def testSuffix(self):
    with temp.AutoTempFilePath(suffix="foo") as filepath:
      self.assertTrue(os.path.exists(filepath))
      self.assertEndsWith(os.path.basename(filepath), "foo")

  def testPrefixAndSuffix(self):
    with temp.AutoTempFilePath(prefix="foo", suffix="bar") as filepath:
      self.assertTrue(os.path.exists(filepath))
      self.assertStartsWith(os.path.basename(filepath), "foo")
      self.assertEndsWith(os.path.basename(filepath), "bar")

  def testDir(self):
    with temp.AutoTempDirPath() as dirpath:
      with temp.AutoTempFilePath(dir=dirpath) as filepath:
        self.assertTrue(os.path.exists(filepath))
        self.assertStartsWith(filepath, dirpath)

  def testRemovesFiles(self):
    with temp.AutoTempFilePath() as filepath:
      self.assertTrue(os.path.exists(filepath))

    self.assertFalse(os.path.exists(filepath))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
from absl.testing import absltest

from grr_response_core.lib.util import text


class AsciifyTest(absltest.TestCase):

  def testAscii(self):
    self.assertEqual(text.Asciify(b"foobar"), "foobar")
    self.assertEqual(text.Asciify(b"+!?#"), "+!?#")

  def testNonAscii(self):
    self.assertEqual(text.Asciify(b"\xff\xfe\xff"), "\\xff\\xfe\\xff")
    self.assertEqual(text.Asciify(b"f\x00\x00bar"), "f\\x00\\x00bar")


class HexifyTest(absltest.TestCase):

  def testEmpty(self):
    self.assertEqual(text.Hexify(b""), "")

  def testEscaped(self):
    self.assertEqual(text.Hexify(b"\xff\xfd\xfa"), "fffdfa")
    self.assertEqual(text.Hexify(b"\x00\x11\x22"), "001122")
    self.assertEqual(text.Hexify(b"\x48\x15\x16\x23\x42"), "4815162342")

  def testAscii(self):
    self.assertEqual(text.Hexify(b"foo"), "666f6f")
    self.assertEqual(text.Hexify(b"bar"), "626172")


class UnescapeStringTest(absltest.TestCase):

  def testWhitespace(self):
    self.assertEqual(text.Unescape("\\n"), "\n")
    self.assertEqual(text.Unescape("\\r"), "\r")

  def testQuotemark(self):
    self.assertEqual(text.Unescape("\\'"), "'")
    self.assertEqual(text.Unescape('\\"'), '"')

  def testMany(self):
    self.assertEqual(text.Unescape("foo\\n\\'bar\\'\nbaz"), "foo\n'bar'\nbaz")


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""This file defines valid configuration contexts."""


from grr_response_core.lib import config_lib

# Different client platforms.
PLATFORM_DARWIN = config_lib.DEFINE_context("Platform:Darwin")
PLATFORM_LINUX = config_lib.DEFINE_context("Platform:Linux")
PLATFORM_WINDOWS = config_lib.DEFINE_context("Platform:Windows")

# Client architectures.
ARCH_AMD64 = config_lib.DEFINE_context("Arch:amd64")
ARCH_I386 = config_lib.DEFINE_context("Arch:i386")
ARCH_PPC64LE = config_lib.DEFINE_context("Arch:ppc64le")
ARCH_AARCH64 = config_lib.DEFINE_context("Arch:aarch64")

# Global system install context - set when GRR is installed globally on the
# system.
GLOBAL_INSTALL_CONTEXT = config_lib.DEFINE_context("Global Install Context")

# Different components.
ADMIN_UI_CONTEXT = config_lib.DEFINE_context("AdminUI Context")
CLIENT_CONTEXT = config_lib.DEFINE_context("Client Context")
CONFIG_UPDATER_CONTEXT = config_lib.DEFINE_context("ConfigUpdater Context")
CONSOLE_CONTEXT = config_lib.DEFINE_context("Console Context")
DEMO_CONTEXT = config_lib.DEFINE_context("Demo Context")
ENTRY_POINT_CONTEXT = config_lib.DEFINE_context("EntryPoint Context")
EXPORT_TOOL_CONTEXT = config_lib.DEFINE_context("ExportTool Context")
HTTP_SERVER_CONTEXT = config_lib.DEFINE_context("HTTPServer Context")
POOL_CLIENT_CONTEXT = config_lib.DEFINE_context("PoolClient Context")
WORKER_CONTEXT = config_lib.DEFINE_context("Worker Context")
FS_FRONTEND_CONTEXT = config_lib.DEFINE_context("FleetspeakFrontend Context")
BENCHMARK_CONTEXT = config_lib.DEFINE_context("Benchmark Context")
GRRAFANA_CONTEXT = config_lib.DEFINE_context("GRRafana Context")

# Client building contexts.
CLIENT_BUILD_CONTEXT = config_lib.DEFINE_context("ClientBuilder Context")
DEBUG_CLIENT_BUILD_CONTEXT = config_lib.DEFINE_context(
    "DebugClientBuild Context")
TARGET_DARWIN = config_lib.DEFINE_context("Target:Darwin")
TARGET_LINUX = config_lib.DEFINE_context("Target:Linux")
TARGET_LINUX_DEB = config_lib.DEFINE_context("Target:LinuxDeb")
TARGET_LINUX_RPM = config_lib.DEFINE_context("Target:LinuxRpm")
TARGET_WINDOWS = config_lib.DEFINE_context("Target:Windows")
TARGET_WINDOWS_MSI = config_lib.DEFINE_context("Target:WindowsMsi")

# Running from the command line.
COMMAND_LINE_CONTEXT = config_lib.DEFINE_context("Commandline Context")

# For debugging.
DEBUG_CONTEXT = config_lib.DEFINE_context("Debug Context")
TEST_CONTEXT = config_lib.DEFINE_context("Test Context")

# Datastores.
MYSQL_DATA_STORE = config_lib.DEFINE_context("MySQLDataStore")

# Client installer context.
INSTALLER_CONTEXT = config_lib.DEFINE_context("Installer Context")



#!/usr/bin/env python
import random
import threading
from unittest import mock

from absl.testing import absltest

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.util import cache
from grr.test_lib import test_lib


class WithLimitedCallFrequencyTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.mock_fn = mock.Mock(wraps=lambda *_: random.random())
    self.mock_fn.__name__ = "foo"  # Expected by functools.wraps.

  def testCallsFunctionEveryTimeWhenMinTimeBetweenCallsZero(self):
    decorated = cache.WithLimitedCallFrequency(rdfvalue.Duration(0))(
        self.mock_fn
    )
    for _ in range(10):
      decorated()

    self.assertEqual(self.mock_fn.call_count, 10)

  def testCacheAlwaysContainsOnlySingleItemWhenMinTimeBetweenCallsZero(self):
    decorated = cache.WithLimitedCallFrequency(rdfvalue.Duration(0))(
        self.mock_fn
    )
    for i in range(10):
      decorated(i)

    self.assertLen(decorated._DebugInternalState()["prev_results"], 1)

  def testCacheIsNotCleanedIfMinTimeBetweenCallsHasNotElapsed(self):
    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)
    for i in range(10):
      decorated(i)

    with test_lib.FakeTime(
        rdfvalue.RDFDatetime.Now()
        + rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    ):
      decorated(11)

    self.assertLen(decorated._DebugInternalState()["prev_results"], 1)

  def testCacheIsCleanedAfterMinTimeBetweenCallsHasElapsed(self):
    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)
    for i in range(10):
      decorated(i)

    self.assertLen(decorated._DebugInternalState()["prev_results"], 10)

  def testCallsFunctionOnceInGivenTimeRangeWhenMinTimeBetweenCallsNonZero(self):
    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)

    now = rdfvalue.RDFDatetime.Now()
    with test_lib.FakeTime(now):
      r1 = decorated()

    with test_lib.FakeTime(now + rdfvalue.Duration.From(15, rdfvalue.SECONDS)):
      r2 = decorated()

    self.assertEqual(r1, r2)
    self.assertEqual(self.mock_fn.call_count, 1)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(30, rdfvalue.SECONDS)):
      r3 = decorated()

    self.assertNotEqual(r1, r3)
    self.assertEqual(self.mock_fn.call_count, 2)

  def testCachingIsDonePerArguments(self):
    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)

    now = rdfvalue.RDFDatetime.Now()
    with test_lib.FakeTime(now):
      r1_a = decorated(1)
      r1_b = decorated(2)

    self.assertNotEqual(r1_a, r1_b)
    self.assertEqual(self.mock_fn.call_count, 2)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(15, rdfvalue.SECONDS)):
      r2_a = decorated(1)
      r2_b = decorated(2)

    self.assertEqual(r1_a, r2_a)
    self.assertEqual(r1_b, r2_b)
    self.assertEqual(self.mock_fn.call_count, 2)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(30, rdfvalue.SECONDS)):
      r3_a = decorated(1)
      r3_b = decorated(2)

    self.assertNotEqual(r1_a, r3_a)
    self.assertNotEqual(r1_b, r3_b)
    self.assertEqual(self.mock_fn.call_count, 4)

  def testDecoratedFunctionIsNotExecutedConcurrently(self):
    event = threading.Event()

    # Can't rely on mock's call_count as it's not thread safe.
    fn_calls = []

    def Fn():
      fn_calls.append(True)
      event.wait()
      return self.mock_fn()

    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(Fn)

    results = []

    def T():
      results.append(decorated())

    threads = []
    for _ in range(10):
      t = threading.Thread(target=T)
      t.start()
      threads.append(t)

    # At this point all threads should be waiting on the function to complete,
    # with only 1 threads actually executing the function. Trigger the event
    # to force that one thread to complete.
    event.set()

    for t in threads:
      t.join()

    self.assertLen(results, len(threads))
    self.assertEqual(set(results), set([results[0]]))
    self.assertLen(fn_calls, 1)

  def testDecoratedFunctionsAreWaitedForPerArguments(self):
    event = threading.Event()

    # Can't rely on mock's call_count as it's not thread safe.
    fn_calls = []

    def Fn(x):
      fn_calls.append(x)
      if x != 42:
        event.wait()
      return x

    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(Fn)

    def T():
      decorated(1)

    t = threading.Thread(target=T)
    t.start()
    try:

      # This should return immediately. There's another function call
      # in progress, but with different arguments, so no locking should occur.
      # I.e. decorated(1) and decorated(42) shouldn't influence each other.
      decorated(42)

    finally:
      event.set()
      t.join()

    self.assertLen(fn_calls, 2)

  def testPropagatesExceptions(self):
    mock_fn = mock.Mock(side_effect=ValueError())
    mock_fn.__name__ = "foo"  # Expected by functools.wraps.

    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    with self.assertRaises(ValueError):
      decorated()

  def testExceptionIsNotCached(self):
    mock_fn = mock.Mock(side_effect=ValueError())
    mock_fn.__name__ = "foo"  # Expected by functools.wraps.

    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    for _ in range(10):
      with self.assertRaises(ValueError):
        decorated()

    self.assertEqual(mock_fn.call_count, 10)

  def testRaisesOnUnhashableArguments(self):
    mock_fn = mock.Mock(side_effect=ValueError())

    decorated = cache.WithLimitedCallFrequency(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    with self.assertRaisesRegex(TypeError, "unhashable type"):
      decorated(dict(foo="bar"))


class WithLimitedCallFrequencyWithoutReturnValueTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.mock_fn = mock.Mock(wraps=lambda *_: None)
    self.mock_fn.__name__ = "foo"  # Expected by functools.wraps.

  def testCallsFunctionEveryTimeWhenMinTimeBetweenCallsZero(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration(0)
    )(self.mock_fn)
    for _ in range(10):
      decorated()

    self.assertEqual(self.mock_fn.call_count, 10)

  def testCacheAlwaysContainsOnlySingleItemWhenMinTimeBetweenCallsZero(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration(0)
    )(self.mock_fn)
    for i in range(10):
      decorated(i)

    self.assertLen(decorated._DebugInternalState()["prev_times"], 1)

  def testCacheIsNotCleanedIfMinTimeBetweenCallsHasNotElapsed(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)
    for i in range(10):
      decorated(i)

    with test_lib.FakeTime(
        rdfvalue.RDFDatetime.Now()
        + rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    ):
      decorated(11)

    self.assertLen(decorated._DebugInternalState()["prev_times"], 1)

  def testCacheIsCleanedAfterMinTimeBetweenCallsHasElapsed(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)
    for i in range(10):
      decorated(i)

    self.assertLen(decorated._DebugInternalState()["prev_times"], 10)

  def testCallsFunctionOnceInGivenTimeRangeWhenMinTimeBetweenCallsNonZero(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)

    now = rdfvalue.RDFDatetime.Now()
    with test_lib.FakeTime(now):
      decorated()

    with test_lib.FakeTime(now + rdfvalue.Duration.From(15, rdfvalue.SECONDS)):
      decorated()

    self.assertEqual(self.mock_fn.call_count, 1)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(30, rdfvalue.SECONDS)):
      decorated()

    self.assertEqual(self.mock_fn.call_count, 2)

  def testCachingIsDonePerArguments(self):
    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(self.mock_fn)

    now = rdfvalue.RDFDatetime.Now()
    with test_lib.FakeTime(now):
      decorated(1)
      decorated(2)

    self.assertEqual(self.mock_fn.call_count, 2)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(15, rdfvalue.SECONDS)):
      decorated(1)
      decorated(2)

    self.assertEqual(self.mock_fn.call_count, 2)

    with test_lib.FakeTime(now + rdfvalue.Duration.From(30, rdfvalue.SECONDS)):
      decorated(1)
      decorated(2)

    self.assertEqual(self.mock_fn.call_count, 4)

  def testDecoratedFunctionIsNotExecutedConcurrently(self):
    event = threading.Event()

    # Can't rely on mock's call_count as it's not thread safe.
    fn_calls = []

    def Fn():
      fn_calls.append(True)
      event.wait()
      return self.mock_fn()

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(Fn)

    results = []

    def T():
      results.append(decorated())

    threads = []
    for _ in range(10):
      t = threading.Thread(target=T)
      t.start()
      threads.append(t)

    # At this point all threads should be waiting on the function to complete,
    # with only 1 threads actually executing the function. Trigger the event
    # to force that one thread to complete.
    event.set()

    for t in threads:
      t.join()

    self.assertLen(results, len(threads))
    self.assertLen(fn_calls, 1)

  def testDecoratedFunctionsAreNotWaitedForPerArguments(self):
    event = threading.Event()

    # Can't rely on mock's call_count as it's not thread safe.
    fn_calls = []

    def Fn(x):
      fn_calls.append(x)
      # Wait if this is the first call.
      if not fn_calls:
        event.wait()

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(Fn)

    def T():
      decorated(1)

    t = threading.Thread(target=T)
    t.start()
    try:
      # This should return immediately, as the wrapped function has no return
      # value and thus doesn't have to block even if the other one is in
      # progress.
      decorated(1)
    finally:
      event.set()
      t.join()

    self.assertLen(fn_calls, 1)

  def testPropagatesExceptions(self):
    mock_fn = mock.Mock(side_effect=ValueError())
    mock_fn.__name__ = "foo"  # Expected by functools.wraps.

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    with self.assertRaises(ValueError):
      decorated()

  def testExceptionIsNotCached(self):
    mock_fn = mock.Mock(side_effect=ValueError())
    mock_fn.__name__ = "foo"  # Expected by functools.wraps.

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    for _ in range(10):
      with self.assertRaises(ValueError):
        decorated()

    self.assertEqual(mock_fn.call_count, 10)

  def testRaisesOnUnhashableArguments(self):
    mock_fn = mock.Mock(side_effect=ValueError())

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    with self.assertRaisesRegex(TypeError, "unhashable type"):
      decorated(dict(foo="bar"))

  def testRaisesIfWrappedFunctionReturnsValue(self):
    mock_fn = mock.Mock(return_value=42)

    decorated = cache.WithLimitedCallFrequencyWithoutReturnValue(
        rdfvalue.Duration.From(30, rdfvalue.SECONDS)
    )(mock_fn)

    with self.assertRaisesRegex(
        AssertionError, "Wrapped function should have no return value"
    ):
      decorated("blah")


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
import contextlib
import io
import sqlite3

from absl.testing import absltest

from grr_response_core.lib.util import sqlite
from grr_response_core.lib.util import temp


class ConnectionContextTest(absltest.TestCase):

  def testQuerySingleColumn(self):
    with temp.AutoTempFilePath(suffix=".sqlite") as db_filepath:
      with contextlib.closing(sqlite3.connect(db_filepath)) as conn:
        with contextlib.closing(conn.cursor()) as cursor:
          cursor.execute("CREATE TABLE foo (bar TEXT)")
          cursor.execute("INSERT INTO foo VALUES ('quux'), ('norf'), ('thud')")

        conn.commit()

        context = sqlite.ConnectionContext(conn)
        results = list(context.Query("SELECT * FROM foo"))
        self.assertEqual(results, [("quux",), ("norf",), ("thud",)])

  def testQueryMultipleColumns(self):
    with temp.AutoTempFilePath(suffix=".sqlite") as db_filepath:
      with contextlib.closing(sqlite3.connect(db_filepath)) as conn:
        with contextlib.closing(conn.cursor()) as cursor:
          cursor.execute("CREATE TABLE foo (bar INTEGER, baz TEXT)")
          cursor.execute("INSERT INTO foo(bar, baz) VALUES (42, 'quux')")
          cursor.execute("INSERT INTO foo(bar, baz) VALUES (108, 'norf')")

        conn.commit()

        context = sqlite.ConnectionContext(conn)
        results = list(context.Query("SELECT bar, baz FROM foo"))
        self.assertEqual(results, [(42, "quux"), (108, "norf")])


class IOConnectionTest(absltest.TestCase):

  def testSmallDatabase(self):
    with temp.AutoTempFilePath(suffix=".sqlite") as db_filepath:
      with contextlib.closing(sqlite3.connect(db_filepath)) as conn:
        with contextlib.closing(conn.cursor()) as cursor:
          cursor.execute("CREATE TABLE foo (bar INTEGER, baz INTEGER)")
          cursor.execute("INSERT INTO foo(bar, baz) VALUES (1, 3), (3, 7)")

        conn.commit()

      with io.open(db_filepath, mode="rb") as db_filedesc:
        db_bytes = db_filedesc.read()

      with sqlite.IOConnection(io.BytesIO(db_bytes)) as context:
        results = list(context.Query("SELECT bar, baz FROM foo"))
        self.assertEqual(results, [(1, 3), (3, 7)])

  def testBigDatabase(self):
    blob = lambda sample: sample * 1024 * 1024

    with temp.AutoTempFilePath(suffix=".sqlite") as db_filepath:
      with contextlib.closing(sqlite3.connect(db_filepath)) as conn:
        with contextlib.closing(conn.cursor()) as cursor:
          cursor.execute("CREATE TABLE foo (bar BLOB)")
          cursor.execute("INSERT INTO foo(bar) VALUES (?)", (blob(b"A"),))
          cursor.execute("INSERT INTO foo(bar) VALUES (?)", (blob(b"B"),))
          cursor.execute("INSERT INTO foo(bar) VALUES (?)", (blob(b"C"),))

        conn.commit()

      with io.open(db_filepath, mode="rb") as db_filedesc:
        with sqlite.IOConnection(db_filedesc) as context:
          results = list(context.Query("SELECT bar FROM foo"))
          self.assertLen(results, 3)
          self.assertEqual(results[0], (blob(b"A"),))
          self.assertEqual(results[1], (blob(b"B"),))
          self.assertEqual(results[2], (blob(b"C"),))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""A module with utility functions for working with SQLite databases."""

from collections.abc import Iterator
import contextlib
import io
import sqlite3
from typing import IO

from grr_response_core.lib.util import temp


class ConnectionContext:
  """A wrapper class around an SQLite connection object.

  This class wraps a low-level SQLite connection that is error-prone and does
  not provide safe context-manager interface.
  """

  def __init__(self, conn: sqlite3.Connection) -> None:
    """Initializes the SQLite connection objects.

    Args:
      conn: A low-level connection to the SQL database.
    """
    self._conn = conn

  def Query(self, query: str) -> Iterator[tuple]:  # pylint: disable=g-bare-generic
    """Queries the underlying database.

    Args:
      query: A query to run.

    Yields:
      Database rows that are results of the query.
    """
    with contextlib.closing(self._conn.cursor()) as cursor:  # pytype: disable=wrong-arg-types
      cursor.execute(query)

      while True:
        rows = cursor.fetchmany(_ROW_FETCH_COUNT)
        if not rows:
          break

        for row in rows:
          yield row


@contextlib.contextmanager
def IOConnection(db_filedesc: IO[bytes]) -> Iterator[ConnectionContext]:
  """A connection to the SQLite database created out of given byte stream.

  Args:
    db_filedesc: A byte stream of the SQLite database file.

  Yields:
    A SQLite connection object that can run queries.
  """
  # SQLite connector can only work with physical database files. Therefore, one
  # needs to dump the contents of the given byte stream to some temporary path.
  with temp.AutoTempFilePath(suffix=".sqlite") as temp_db_filepath:
    with io.open(temp_db_filepath, mode="wb") as temp_db_filedesc:
      _CopyIO(input=db_filedesc, output=temp_db_filedesc)

    with contextlib.closing(sqlite3.connect(temp_db_filepath)) as conn:  # pytype: disable=wrong-arg-types
      yield ConnectionContext(conn)


def _CopyIO(input: IO[bytes], output: IO[bytes]) -> None:  # pylint: disable=redefined-builtin
  """Copies contents of one binary stream into another.

  Args:
    input: An input stream to read from.
    output: An output stream to write to.
  """
  while True:
    buf = input.read(_FILE_BUFFER_SIZE)
    if not buf:
      break

    output.write(buf)


_FILE_BUFFER_SIZE = 1024 * 1024  # 1 MiB.
_ROW_FETCH_COUNT = 1024



#!/usr/bin/env python
import gzip
import io
import os
import struct

from absl.testing import absltest

from grr_response_core.lib.util import chunked
from grr_response_core.lib.util import gzchunked


class SerializeTest(absltest.TestCase):

  def testEmpty(self):
    serialized = list(gzchunked.Serialize(iter([])))
    self.assertEmpty(serialized)

  def testSingleEntry(self):
    data = [b"foo"]

    serialized = list(gzchunked.Serialize(iter(data), chunk_size=1024))
    self.assertLen(serialized, 1)

  def testMultipleSmallEntries(self):
    data = [b"foo", b"bar", b"baz", b"quux"]

    serialized = list(gzchunked.Serialize(iter(data), chunk_size=1024))
    self.assertLen(serialized, 1)

  def testMultipleBigEntries(self):
    data = [os.urandom(1024 * 1024) for _ in range(8)]

    serialized = list(gzchunked.Serialize(iter(data), chunk_size=(1024 * 1024)))
    self.assertGreater(len(serialized), 1)
    self.assertLessEqual(len(serialized), len(data))


class DeserializeTest(absltest.TestCase):

  def testIncorrectSize(self):
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode="wb") as filedesc:
      filedesc.write(struct.pack("!I", 42))

    with self.assertRaises(chunked.IncorrectSizeTagError):
      list(gzchunked.Deserialize(iter([buf.getvalue()])))

  def testIncorrectData(self):
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode="wb") as filedesc:
      filedesc.write(struct.pack("!Q", 8))
      filedesc.write(b"quux")

    with self.assertRaises(chunked.ChunkTruncatedError):
      list(gzchunked.Deserialize(iter([buf.getvalue()])))

  def testEmpty(self):
    serialized = list(gzchunked.Serialize(iter([])))
    deserialized = list(gzchunked.Deserialize(iter(serialized)))

    self.assertEqual(deserialized, [])

  def testSingleEntry(self):
    data = [b"foo"]

    serialized = list(gzchunked.Serialize(iter(data)))
    deserialized = list(gzchunked.Deserialize(iter(serialized)))

    self.assertEqual(deserialized, data)

  def testMultipleEntries(self):
    data = [b"foo", b"bar", b"baz", b"quux", b"norf", b"thud"]

    serialized = list(gzchunked.Serialize(iter(data)))
    deserialized = list(gzchunked.Deserialize(iter(serialized)))

    self.assertEqual(deserialized, data)

  def testEmptyData(self):
    data = [b"", b"", b""]

    serialized = list(gzchunked.Serialize(iter(data)))
    deserialized = list(gzchunked.Deserialize(iter(serialized)))

    self.assertEqual(deserialized, data)

  def testNoChunks(self):
    deserialized = list(gzchunked.Deserialize(iter([])))
    self.assertEmpty(deserialized)

  def testMultipleChunks(self):
    data = [os.urandom(1024 * 1024) for _ in range(8)]

    serialized = list(gzchunked.Serialize(iter(data), chunk_size=(1024 * 1024)))
    self.assertGreater(len(serialized), 1)

    deserialized = list(gzchunked.Deserialize(iter(serialized)))
    self.assertEqual(deserialized, data)


if __name__ == "__main__":
  absltest.main()
